SystemOrganization addCategory: #'Prolog-Compiler'!


Decompiler subclass: #LDecompiler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Compiler'!

Compiler subclass: #LCompiler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Compiler'!

LDecompiler class
	instanceVariableNames: ''!

Parser subclass: #LParser
	instanceVariableNames: 'prologMethodSelector prologMethodArgument cuts temporaries statements smallFlag'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Compiler'!

LParser subclass: #LParserForTesting
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Compiler'!

LParser class
	instanceVariableNames: ''!

LParserForTesting class
	instanceVariableNames: ''!

LCompiler class
	instanceVariableNames: ''!


!LCompiler commentStamp: 'spfa 6/20/2008 12:39' prior: 0!
Compiler for Prolog class tree!
]style[(30)!

!LDecompiler commentStamp: 'spfa 2/26/2017 13:04' prior: 0!
Decompiler for Prolog class tree

Does not produce actual working code for Prolog methods because of the selector dispatching system; see usages of #selectorForPredicate:!
]style[(170)!

!LParser commentStamp: 'bolot 12/13/2002 15:14' prior: 0!
RE: safeGuardReturnBlock
Makes sure that the method returns, even if no match occurs.

RE: queryUndefined
Disabled for Prolog methods (because undefined Prolog variables are unbounds).

RE: prologMethodSelector
Can be used to determine if this is a smalltalk method or prolog method.

RE: Smalltalk guys
Strings, symbols, numbers, and literal points (45@65) are recognized.

RE: evaluation
Support for nicer, more Prolog-like syntax for evaluation is done through direct use of LCompiler.
Example: LCompiler evaluate: 'Family new :? father(''John'', x).'
(see Family comment for current situation)

This class and other related classes in this category were originally written by or are based on the work by Mike Teng <miket@objectshare.com>, then at Digitalk, as part of the Prolog/V implementation distributed with Smalltalk V/286 in 1988.
We gratefully acknowledge Mike's permission to port his code to Squeak.

The port to Squeak is the work of Bolot Kerimbaev <bolot@cc.gatech.edu>, a summer intern at Walt Disney Imagineering.!
]style[(1032)!

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/3/2022 23:59'!
addTemp: aTerm
	"Decide whether aTerm should be declared as a temporay."
	| var term |

	self assert: aTerm first isUppercase.
	term := aTerm withFirstCharacterDownshifted.
	
	var := encoder encodeVariable: term
		ifUnknown: [encoder bindTemp: term].

	(temporaries includes: var) ifTrue: [^var].

	temporaries add: var.
	statements addFirst: 
		(AssignmentNode new
			variable: var
			value: (self send: #newTerm: to: 'LogicRef' with: {aTerm})).
	^var! !

!LParserForTesting methodsFor: 'as yet unclassified' stamp: 'spfa 10/20/2022 13:29'!
addTemp: aTerm

	^ aTerm! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
argumentList: aList
	^ aList isConsCell
		ifTrue: [self arguments: aList]
		ifFalse: [aList]! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/20/2022 11:27'!
arguments: aList
	"Answer a MessageNode for predicate arguments."
	| receiver bList |

	aList ifNil: [^ nil] "zero argument".

	receiver := self send: #car:
			to: (aList isRelation ifTrue: ['Relation'] ifFalse: ['PrologList'])
			with: {self argumentList: aList head}.

	aList cdr ifNil: [^ receiver] "single argument". 

	aList cdr isConsCell ifFalse: ["[x | y]"
		^ self send: #car:cdr:
			to: (aList isRelation ifTrue: ['Relation'] ifFalse: ['PrologList'])
			with: {self argumentList: aList head . self argumentList: aList tail}].

	"more than one arguments"
	bList := aList.
	[(bList := bList tail) isConsCell] whileTrue:
		[receiver := self send: #addLast:
						to: receiver
						with: {self argumentList: bList head}].
	bList ifNotNil:
		[^ self send: #append:
					to: receiver
					with: {self argumentList: bList}].
	^ receiver! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/17/2022 11:58'!
bracedTerms
	
	self notYetImplemented! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/12/2022 15:30'!
buildPredicateFrom: anArray

	^ self buildPredicateFrom: anArray ensurePrecedenceBelow: nil! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/14/2022 15:51'!
buildPredicateFrom: anArray ensurePrecedenceBelow: aNumber

	| topop ti prec |

	anArray size = 1 ifTrue: [^ anArray anyOne].

	prec := -1.
	anArray reverseDoWithIndex: [:ea :n |
		(ea class == PrologOperator 
			and: [ea precedence > prec])
		ifTrue: [
			ti := n.
			topop := ea.
			prec := topop precedence]].

	(aNumber notNil and: [prec = aNumber] and: [topop type first = $x])
		ifTrue: [self error: 'Ambiguous operator precedence - parens required']. 

	(topop type = #xfx) ifTrue: [
		^ {topop operator. 
			self buildPredicateFrom: (anArray copyFrom: 1 to: ti-1)
					ensurePrecedenceBelow: prec. 
			self buildPredicateFrom: (anArray copyFrom: ti+1 to: anArray size)
					ensurePrecedenceBelow: prec} 
				asPrologRelation].

	(topop type = #yfx) ifTrue: [
		^ {topop operator. 
			self buildPredicateFrom: (anArray copyFrom: 1 to: ti-1)
					ensurePrecedenceBelow: prec. 
			self buildPredicateFrom: (anArray copyFrom: ti+1 to: anArray size)} 
				asPrologRelation].

	(topop type = #xfy) ifTrue: [
		^ {topop operator. 
			self buildPredicateFrom: (anArray copyFrom: 1 to: ti-1). 
			self buildPredicateFrom: (anArray copyFrom: ti+1 to: anArray size)
					ensurePrecedenceBelow: prec} 
				asPrologRelation].

	(#(fx fy) includes: topop type) ifTrue: [
		^ {topop operator. 
			self buildPredicateFrom: (anArray copyFrom: ti+1 to: anArray size)} 
				asPrologRelation].

	(#(xf yf) includes: topop type) ifTrue: [
		^ {topop operator. 
			self buildPredicateFrom: (anArray copyFrom: 1 to: ti-1)} 
				asPrologRelation].

	self error: 'Badly defined operator: ', topop operator! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/17/2022 17:40'!
clauseDelimiters

	^ #(period rightParenthesis rightBracket verticalBar doIt)
! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/20/2022 10:52'!
commaTreeOf: anObject

	| array |

	(anObject isPrologList and: [anObject car == #',']) 
		ifTrue: [^ (self commaTreeOf: anObject second), 
					(self commaTreeOf: anObject third)].

	(array := anObject) isArray ifFalse: [^ {anObject}].

	^ Array streamContents: [:str |
		array do: [:ea | 
			str nextPutAll: (self commaTreeOf: ea)]]! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/12/2022 00:31'!
compileDCG: aPrologList

	"Definite clause grammar 
	--> syntactic sugar"

	| block relations tmps tmp term args terms |

	(self match: #period) ifFalse: [self expected: 'Period']. 

	args := aPrologList head isSymbol 
				ifTrue: [#()] 
				ifFalse: [aPrologList head cdr topForms].

	terms := aPrologList tail.

	relations := Array streamContents: [:str |

		terms topForms do: [:t |
			term := t.
			term isSymbol ifTrue: [term := Relation car: term].
			term ifNil: [str nextPut: (#terminal -> nil)]
				ifNotNil: [
					(term isPrologList and: [term isRelation not]) 
						ifTrue: [
							term topForms "terminals" do: [:form |
								str nextPut: (#terminal -> form)]] 
						ifFalse: [str nextPut: term]].]].

	tmps := (1 to: relations size) collect: [:n |
		self addTemp: 'S', (n-1)].

	tmp := self addTemp: 'S'.

	relations := relations collectWithIndex: [:ea :n |
		| s0 s1 |
		n = relations size 
			ifTrue: [s0 := tmps last. s1 := tmp]
			ifFalse: [s0 := tmps at: n. s1 := tmps at: n+1].
		ea class == Association "terminals"
			ifTrue: [
				(ea value isNil and: [args isEmpty]) 
					ifTrue: [{#unify . s0.  s1} asPrologRelation]
					ifFalse: [{#cccc . s0. ea value . s1} asPrologRelation]] 
			ifFalse: [(ea topForms, {s0. s1}) asPrologRelation]].

	block := self predicateSeriesFromRelations: relations.

	^ self send: #unify:with:then:
		to: 'self'
		with: {
				self send: #key to: prologMethodArgument.
				self arguments: (args, {tmps first. tmp}) asPrologList.
				(hereType == #doIt and: [cuts not])
						ifTrue: [self returnBlock: block] 
						ifFalse: block
				}! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/20/2022 11:11'!
compileRule: predicate block: aBlockNode 
"
	predicate head == prologMethodSelector
		ifFalse: [self notify: 'predicate name different'].
"
	(self match: #period) ifFalse: [self expected: 'Period']. 

	^ self send: #unify:with:then:
		to: 'self'
		with: {
				self send: #key to: prologMethodArgument.
				self arguments: predicate tail.
				(hereType == #doIt and: [cuts not])
						ifTrue: [self returnBlock: aBlockNode]  "optimize the stack" 
						ifFalse: aBlockNode
				}! !

!LParserForTesting methodsFor: 'as yet unclassified' stamp: 'spfa 10/16/2022 13:20'!
compileRule: predicate block: aBlockNode 

	(self match: #period) ifFalse: [self expected: 'Period']. 
! !

!LParserForTesting methodsFor: 'as yet unclassified' stamp: 'spfa 10/20/2022 13:28'!
compiling: aString

	| cue myStream s p subSelection |

	cue := (CompilationCue
			source: aString
			context: nil
			receiver: nil
			class: Prolog
			environment: UndefinedObject environment
			requestor: nil).

	self setCue: cue.

	myStream := cue sourceStream.

	 p := myStream position.
	 s := myStream upToEnd.
	 myStream position: p.

	 doitFlag := false.
"	 (self encoderFromCue: cue) init: cue notifying: self."
	 self init: myStream cue: cue failBlock: ["self halt"].

	 subSelection := self interactive and: [cue requestor selectionInterval = (p + 1 to: p + s size)].

    temporaries := OrderedCollection new.
    statements := OrderedCollection new.
    typeTable := typeTable copy.

	^ self! !

!LCompiler class methodsFor: 'evaluating' stamp: 'spfa 10/13/2022 16:11'!
deResultify: anArray

	^ anArray select: [:ea | ea class ~~ LogicUnification] thenCollect: #logicResult! !

!LDecompiler methodsFor: 'as yet unclassified' stamp: 'spfa 9/29/2022 00:42'!
decompile: selector in: aClass method: aMethod using: aConstructor

"wrong - many non-predicate selectors starting the prologKeyword are indeed unary
 #logicValue, etc"
[
	((selector beginsWith: Logic prologKeyword)
		and: [selector isUnary])
	"restore the one-argument form of selector"
	ifTrue: [^ super decompile: (selector, ':') asSymbol in: aClass method: aMethod using: aConstructor].
].

	^ super decompile: selector in: aClass method: aMethod using: aConstructor

! !

!LCompiler class methodsFor: 'accessing' stamp: 'spfa 11/18/2009 18:10'!
decompilerClass

	^LDecompiler! !

!LParser methodsFor: 'private' stamp: 'spfa 10/20/2022 13:24'!
encoderFromCue: aCompilationCue
	^encoder ifNil:
		[encoder := (aCompilationCue encoderClass ifNil: [CompiledMethod preferredBytecodeSetEncoderClass]) new]! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/19/2022 23:16'!
ensureCallable: aPrologRelation

	| prolog |

	prolog := cue getClass. "would be better if the parser had a reference to the expert"

	aPrologRelation car: (Logic makeLogicSelector: 
		((prolog binaryPredicates keys includes: aPrologRelation car)
			ifTrue: [prolog selectorForBinaryPredicate: aPrologRelation car]
			ifFalse: [aPrologRelation car])).

	^ aPrologRelation! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/20/2022 11:45'!
expandTerm: term

	"no-op at the moment"

	^ term! !

!LCompiler class methodsFor: 'evaluating' stamp: 'spfa 10/13/2022 15:43'!
generatorOnQueryTo: aLogic for: eString withBindings: aBoolean

	^ Generator on: [:g |
		[self evaluate: 'self yourself :? ', eString for: aLogic logged: false]
			on: PrologAnswerStep 
			do: [:ex | 
				g yield: (aBoolean ifTrue: [ex answer] 
					ifFalse: [self deResultify: ex answer]).
				ex resume]]! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/20/2022 11:29'!
goalSeries
	"Do the work for method statements by
	 adding each goal to statements."
	" (Prolog new :? goal1(x,y), goal2(z,y)) do: [:i | ...] "

	^ self arguments: self readArguments! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/20/2022 13:09'!
initialize

	super initialize.
	typeTable := TypeTable copy "safer".
	smallFlag := false! !

!LParser methodsFor: 'error correction' stamp: 'spfa 11/18/2009 18:10'!
interactive

	^ (requestor isKindOf: REPLMorph) not
			and: [super interactive]! !

!LParser methodsFor: 'scanning' stamp: 'spfa 10/15/2022 17:48'!
matchAnyOf: types

	(types includes: hereType)
		ifTrue: 
			[self advance.
			^true].
	^false! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/17/2022 21:44'!
matchCases: anArray

	anArray do: [:assoc |
		(self match: assoc key) ifTrue: [^ assoc value value]].

	PrologNoMatch signal! !

!LParser methodsFor: 'expression types' stamp: 'spfa 10/13/2022 00:24'!
messagePart: level repeat: repeat
	"bolot 8/5/1999 18:14"
	"This is the same as Parser>>messagePart:repeat:"
	"with only ONE addition: support for #:? for calling Prolog stuff"

	| start receiver selector args precedence words keywordStart |
	[receiver := parseNode.
	(hereType == #keyword and: [level >= 3])
		ifTrue: 
			[start := self startOfNextToken.
			selector := WriteStream on: (String new: 32).
			args := OrderedCollection new.
			words := OrderedCollection new.
			[hereType == #keyword]
				whileTrue: 
					[keywordStart := self startOfNextToken + requestorOffset.
					selector nextPutAll: self advance.
					words addLast: (keywordStart to: self endOfLastToken + requestorOffset).
					self primaryExpression ifFalse: [^self expected: 'Argument'].
					self messagePart: 2 repeat: true.
					args addLast: parseNode].
			(Symbol hasInterned: selector contents ifTrue: [ :sym | selector := sym])
				ifFalse: [ selector := self correctSelector: selector contents
										wordIntervals: words
										exprInterval: (start to: self endOfLastToken)
										ifAbort: [ ^ self fail ] ].
			precedence := 3]
		ifFalse: [((hereType == #binary or: [hereType == #verticalBar])
				and: [level >= 2])
				ifTrue: 
					[start := self startOfNextToken.
					selector := self advance asSymbol.
					self primaryExpression ifFalse: [^self expected: 'Argument'].
					self messagePart: 1 repeat: true.
					args := Array with: parseNode.
					precedence := 2]
				ifFalse: [hereType == #word
						ifTrue: 
							[start := self startOfNextToken.
							selector := self advance.
							args := #().
							words := OrderedCollection with:
									(start + requestorOffset to: self endOfLastToken + requestorOffset).
							(Symbol hasInterned: selector ifTrue: [ :sym | selector := sym])
								ifFalse: [ selector := self correctSelector: selector
													wordIntervals: words
													exprInterval: (start to: self endOfLastToken)
													ifAbort: [ ^ self fail ] ].
							precedence := 1]
						ifFalse: [(hereType == #colonQuestion)
								ifTrue: [selector := #inferSmalltalk:.
										"change typeTable for Prologging"
										typeTable at: $, asciiValue put: #comma;
											at: $!! asciiValue put: #exclamationMark.
										self advance. 
										args := Array with: self goalSeries .
										"restore typeTable for Smalltalking"
										typeTable at: $, asciiValue put: (TypeTable at: $, asciiValue);
											at: $!! asciiValue put: (TypeTable at: $!! asciiValue).
										precedence := 3 "TODO: pick right precedence"]
								ifFalse: [^args notNil]]]].
	parseNode := MessageNode new
				receiver: receiver
				selector: selector
				arguments: args
				precedence: precedence
				from: encoder
				sourceRange: (start to: self endOfLastToken).
	repeat]
		whileTrue: [].
	^true! !

!LParser methodsFor: 'expression types' stamp: 'spfa 10/11/2022 23:15'!
method: doit context: ctxt

    | sap blk prim messageComment methodNode |

    "need temporaries/statements in doits - to enable unbounds"
    temporaries := OrderedCollection new.
    statements := OrderedCollection new.
    "may need to change (temporarily) for doits"
    typeTable := typeTable copy.

    "check if it's a Prolog method definition"
	((doit not and: [here beginsWith: Logic prologKeyword]) 
		and: [(here last == $:) not])
     ifTrue:
            [^self prologMethod: doit context: ctxt].

    sap := self pattern: doit inContext: ctxt.
    "sap={selector, arguments, precedence}"
  "  (sap at: 2) do: [:argNode | argNode beMethodArg]."   "... or   isArg: true"
    temporaries := self temporaries asOrderedCollection.

    messageComment := currentComment.
    currentComment := nil.
    prim := doit ifTrue: [0] ifFalse: [self primitive].
    self statements: #() innerBlock: doit.

    parseNode statements: statements , parseNode statements.

    blk := parseNode.
    doit ifTrue: [blk returnLast]
        ifFalse: [blk returnSelfIfNoOther].
    (self termDelimiters includes: hereType) ifFalse: [^self expected: 'Nothing more'].
 "   self interactive ifTrue: [self removeUnusedTemps]."

    methodNode := self newMethodNode comment: messageComment.
    ^methodNode
        selector: (sap at: 1)
        arguments: (sap at: 2)
        precedence: (sap at: 3)
        temporaries: temporaries
        block: blk
        encoder: encoder
        primitive: prim! !

!LParser methodsFor: 'expression types' stamp: 'spfa 4/15/2010 17:57'!
method: doit context: ctxt encoder: encoderToUse
	" pattern [ | temporaries ] block => MethodNode."

	encoder := encoderToUse.
	^self method: doit context: ctxt! !

!LParserForTesting methodsFor: 'as yet unclassified' stamp: 'spfa 10/16/2022 12:07'!
parseCue: aCue noPattern: noPattern ifFail: aBlock 
	"Just initialize the parser"

	| myStream s p subSelection |
	myStream := aCue sourceStream.

	 p := myStream position.
	 s := myStream upToEnd.
	 myStream position: p.

	 doitFlag := noPattern.
	 (self encoderFromCue: aCue) init: aCue notifying: self.
	 self init: myStream cue: aCue failBlock: [^aBlock value].

	 subSelection := self interactive and: [cue requestor selectionInterval = (p + 1 to: p + s size)].

	encoder := self encoder class new.! !

!LCompiler class methodsFor: 'accessing' stamp: 'bolot 7/2/1999 17:25'!
parserClass
	"Return a parser class to use for parsing method headers."

	^LParser! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/19/2022 23:15'!
predCut
	cuts := true.
	^ PrologList with: #cut 
		with: (self returnBlock: (self send: #asArray to: 'answer'))
! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/19/2022 14:46'!
predicateSeries: aCons
	| pred |
	pred := self ensureCallable: aCons car asPrologRelation.
    ^BlockNode new
		arguments: #()
		statements: {self send: pred head
							to: 'self'
							with: {self send: #key:value:
										to: 'Association'
										with: {
											self arguments: pred tail .
											aCons cdr 
												ifNotNil: [self predicateSeries: aCons cdr]
												ifNil: [self send: #value
															to: prologMethodArgument
															with: #()]}}}
		returns: false
		from: encoder! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/7/2022 23:56'!
predicateSeriesFromRelations: anArray
	| pred |
	pred := anArray first.

    ^BlockNode new
		arguments: #()
		statements: {
			self send: (Logic makeLogicSelector: pred head)
				to: 'self'
				with: {self send: #key:value:
							to: 'Association'
							with: {
								self arguments: pred tail .
								anArray size > 1 
									ifTrue: [self predicateSeriesFromRelations: anArray allButFirst]
									ifFalse:[self send: #value
												to: prologMethodArgument
												with: #()]}}}
		returns: false
		from: encoder! !

!LParser methodsFor: 'hack' stamp: 'spfa 10/19/2022 13:52'!
primitive

	^ self pragmaPrimitives! !

!LParser methodsFor: 'expression types' stamp: 'spfa 11/18/2009 18:10'!
prologMethod: doit context: ctxt

	| blk messageComment methodNode blockComment |

	prologMethodSelector := (self pattern: doit inContext: ctxt) first.

	typeTable at: $, asciiValue put: #comma.
	typeTable at: $!! asciiValue put: #exclamationMark.

	(prologMethodArgument := encoder bindArg: #assoc).
	" prologMethodArgument  beMethodArg. "  "... or   isArg: true"
	messageComment := currentComment.
	currentComment := nil.
	" self statements: #() innerBlock: doit. "
	parseNode := BlockNode new
				arguments: #()
				statements: (self ruleSeries, self safeGuardReturnStatements)
				returns: doit
				from: encoder.
	blockComment := currentComment.
	currentComment := nil.
	parseNode comment: blockComment.
	blk := parseNode.
	hereType == #doIt ifFalse: [^self expected: 'Nothing more'].
	methodNode := MethodNode new comment: messageComment.

	^methodNode
		selector: prologMethodSelector
		arguments: {prologMethodArgument}
		precedence: 3
		temporaries: temporaries
		block: blk
		encoder: encoder
		primitive: 0! !

!LParser methodsFor: 'error correction' stamp: 'spfa 9/27/2022 21:26'!
queryUndefined
	"a lot of Prolog variables are undefined because they are unbound"

[
	"let's prevent unnecessary interrogation (unless it's a Smalltalk method)"
	prologMethodSelector ifNil: [^super queryUndefined]
]! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/20/2022 11:29'!
readArguments

	| term aList |

	term := self readTerm.
	aList := PrologList car: term.
	[
		(self match: #comma)
			ifTrue: [aList addLast: self readTerm]
			ifFalse: [^aList].
	] repeat! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/20/2022 00:18'!
readClause

	^ self readToken: #all delimiters: self clauseDelimiters! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/15/2022 17:44'!
readList
	"Answer a list of predicate terms."
	| term aList |

	(self match: #rightBracket) ifTrue: [^ nil].
	term := self readTerm.
	aList := PrologList car: term.
	[
		(self match: #verticalBar) ifTrue: [
				aList append: self readTerm.
				(self match: #rightBracket) ifTrue: [^ aList].
				self expected: 'Right bracket'].
		(self match: #comma)
			ifTrue: [aList addLast: self readTerm]
			ifFalse: [
				(self match: #rightBracket) ifTrue: [^aList].
				self expected: 'Right bracket']
	] repeat! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/20/2022 13:09'!
readPredicate: aSymbol

	| result |

	self advance.

	(aSymbol = 'iss' or: [aSymbol = 'consult'])
		ifFalse: [^ self structure: aSymbol].

	"special cases: involve parsing smalltalk code"

	result := Relation with: aSymbol.
	result addLast: self readTerm. "the Prolog side, first argument"

	smallFlag := true.

	(self match: #comma)
		ifFalse: [self expected: 'Term (comma?)'].
	self expression
		ifFalse: [self expected: 'Expression'].
    (self match: #rightParenthesis)
		ifFalse: [self expected: 'Right parenthesis'].
	result addLast:
		(BlockNode new
			arguments: #()
			statements: {parseNode}
			returns: false
			from: encoder).

	smallFlag := false.

	^result! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/20/2022 11:50'!
readTerm

	^ self expandTerm: (self readToken: #single delimiters: self termDelimiters)! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/20/2022 00:56'!
readToken

	| item hend fragment |

	hend := hereEnd. "used to detect a gap before a parens"
	item := here.

	[fragment := self matchCases: {

		"BRACED - {...term1, term2, etc..}"
		#leftBrace -> [self bracedTerms].

		"LIST - fragment is a list of terms"
		#leftBracket -> [self readList].

		"AND as (...term1, term2, etc...)"
		#leftParenthesis -> [ | f |
				f := self readClause.
				(self match: #rightParenthesis) 
					ifTrue: [f]
					ifFalse: [self expected: 'right parenthesis']]. 

		#comma -> [PrologOperator named: #,.]. 
		#colonDash -> [PrologOperator named: #':-']. 

		"ATOMIC - fragment is a number, string or literal"
		#string -> [item].
		#number -> [item].
		#literal -> [item].

		"DON'T CARE VARIABLE - fragment is an underscore"
		#leftArrow "should be #underscore" -> [self send: #new to: 'LogicDontCare'].
	}

	] on: PrologNoMatch
	do: [
		"here we deal with a symbol: either a variable, an plain atom, or a functor"
		(self matchAnyOf: #(word binary)) ifFalse: [self expected: 'Symbol'].

		"VARIABLE - fragment is a named variable"
		(item first isLetter and: [item first isUppercase]) ifTrue: [^ self addTemp: item].

		(hereType = #leftParenthesis) ifTrue: [

				(hereMark - hend > 1) ifFalse: ["no gap between symbol and $(, so a functor"
					^ self readPredicate: item asSymbol].
	
				"a gap, so an operator"
				^ (PrologOperator named: item asSymbol) ifNil: [self expected: 'Operator']].

		^ PrologOperator ifAnyFor: item].

	^ fragment

		! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/19/2022 23:21'!
readToken: parsingMode delimiters: delimiters

	"This mid-level, recursive, utility method either returns 
		- one or more Prolog terms,
		- an array or terms and operators to be analysed and structured according to their associativities and precedences into a clause or a term by #buildPredicateFrom:
		- a PrologList of terms articulated by commas, to be converted into a clause or term by #commaTreeOf:.

	What it does exactly depends on the interplay of three different parsing modes, and two different sets of delimiters. The algorithm is the distillation of heuristical incremental test-driven progress - to be honest, why and how it works is not crystal clear to me (the author, spfa).

	1) parsing modes:
		#all				read all upcoming terms		(return one or more terms, a clause)
		#single			read the current term only		(return a term)
		#operators		read all, considers comma and dashColon as terms themselves
														(return an Array of terms and operators)

	Basically we can parse a clause with #all when there are no operators; that is, no functors appearing in infix, postfix or prefix notation, and no --> (the syntactic sugar for definite clause grammars aka DCG).

	So all high-level senders just use #all or #single. It is only when, during the parsing, a term is found that does not end the clause as expected, that the algorithm switches to #operators.

	Note that the #all mode is not only meant for reading top-level clauses, but also for reading arguments and terms appearing within parenthesis. Indeed some of the recursive calls appear in #readToken, via higher-level senders.

	2) delimiters:

		term delimiters are
			#(rightParenthesis rightBracket verticalBar doIt period comma)

		clause delimiters are
			#(rightParenthesis rightBracket verticalBar doIt period) 

	Because the method is recursive, the so-called 'term' and 'clause' sets of delimiters also include the delimiters ending a subexpression, that is #(rightParenthesis rightBracket verticalBar)"


	(self match: #exclamationMark) ifTrue: [^ self predCut]. 

	(delimiters includes: hereType)	ifTrue: [self expected: 'Term'].

	self readToken in: [:tok |
		| delim |
		(delimiters includes: hereType) ifTrue: [^ tok].
		parsingMode = #operators ifTrue: [^ tok].
		delim := parsingMode = #all 
					ifTrue: [self clauseDelimiters] 
					ifFalse: [self termDelimiters].
		^ self buildPredicateFrom: 
			(Array streamContents: [:str |
				str nextPut: tok.
				[
					str nextPut: 
						(self readToken: #operators delimiters: self clauseDelimiters).
					delim includes: hereType
				] whileFalse])]
! !

!LParserForTesting class methodsFor: 'as yet unclassified' stamp: 'spfa 10/16/2022 12:07'!
reading: aString

	^ self new compiling: aString readStream! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
returnBlock: anExpression
	"Answer a Block which returns anExpression."

	^BlockNode new
		arguments: #()
		statements: {
			ReturnNode new
				expr: anExpression
				encoder: encoder
				sourceRange: (1 to: 1)}
		returns: true
		from: encoder! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/20/2022 12:06'!
rule
	"rule:
		predicate [:- predicates] 
	or
		head --> body"

	| clause |

	cuts := false.

	clause := self readClause.

	"case 1) predicate :- predicates."
	(clause head = #':-') ifTrue: [
		clause := (self commaTreeOf: clause tail topForms) asPrologList.
		^ self compileRule: clause head asPrologRelation
				block: (self predicateSeries: clause tail)].

	"case 2) head --> body."
	(clause head = #-->) ifTrue: [
		^ self compileDCG: (self commaTreeOf: clause tail topForms) asPrologList].

	"case 3) predicate." 	"(alone)"
	^ self compileRule: clause
			block: (self send: #value to: prologMethodArgument)
! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
ruleSeries
	"Do the work for method statements by
	 adding each rule to statements."

	[statements add: self rule. source atEnd] whileFalse.
	^statements! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
safeGuardReturnStatements
	^ {
		self send: #ifTrue:ifFalse: 
			to: (self send: #isEmptyOrNil to: 'answer')
			with: {
					self returnBlock: (encoder encodeVariable: 'nil') .
					self returnBlock: (self send: #asArray to: 'answer')
					}
		}
	! !

!LDecompiler methodsFor: 'as yet unclassified' stamp: 'spfa 9/29/2022 01:24'!
send: selector super: superFlag numArgs: numArgs

	| sel |

	sel := selector.

	(selector isUnary
		and: [numArgs > 0]
		and: [selector beginsWith: Logic prologKeyword])
	ifTrue: [
	"here selector is a Prolog predicate that will be dynamically looked up in Logic>>#doesNotUnderstand. 	we 'decompile' it by giving an abstract selector, for information only."
		sel := ('logic.', 
				(selector allButFirst: Logic prologKeyword size) 					withFirstCharacterDownshifted,
				 ':') asSymbol].

	^ super send: sel super: superFlag numArgs: numArgs
! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
send: selector to: receiver 

	^ self send: selector to: receiver with: #()! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
send: selector to: receiver with: arguments
	"selector is a Symbol or a Selector
	node. receiver can be
	either a String or other nodes."

	^ MessageNode new
		receiver: (receiver isString
						ifTrue: [encoder encodeVariable: receiver]
						ifFalse: [receiver])
		selector: (selector asSymbol)
		arguments: (arguments collect: [:elem |
							(elem isKindOf: ParseNode)
								ifTrue: [elem]
								ifFalse: [encoder encodeLiteral: elem]])
		precedence: 3
		from: encoder
		sourceRange: (1 to: 1)! !

!LParserForTesting methodsFor: 'as yet unclassified' stamp: 'spfa 10/16/2022 13:19'!
send: selector to: receiver with: arguments! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/14/2022 11:59'!
structure: aSymbol
	"Process the argument part of a structure
	 whose functor is aSymbol. Current token
	 is now the left paren. Answer a Relation
	 representing the structure."

	| result |

	result := Relation with: aSymbol.
	(self match: #rightParenthesis) ifTrue: [^ result].

	[result addLast: self readTerm. (self match: #comma)] whileTrue.

	(self match: #rightParenthesis)
		ifFalse: [self expected: 'Right parenthesis'].

	^result! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/19/2022 23:20'!
termDelimiters

	^ #(period rightParenthesis rightBracket verticalBar comma doIt)
! !

!LParserForTesting methodsFor: 'as yet unclassified' stamp: 'spfa 10/16/2022 11:12'!
token

	^ token! !

!LParser methodsFor: 'multi-character scans' stamp: 'spfa 10/20/2022 13:10'!
xBinary

	| startOfToken |
	tokenType := #binary.
	startOfToken := mark.

	token := String with: self step.

	smallFlag ifFalse: ["HACK"
		"Allow the proper parsing of negative numbers"
		(token = '-'  
			and: [hereType ~= #number] 
			and: [(self typeTableAt: hereChar) == #xDigit])
		ifTrue: [
			self xDigit. token := token negated. ^ self]].

	[(self typeTableAt: hereChar) == #xBinary or: [(self typeTableAt: hereChar) == #verticalBar]] whileTrue:
		[(hereChar == $- and: [(self typeTableAt: aheadChar) == #xDigit])
			ifTrue: [^self ambiguousSelector: (token , '-')
					inRange: (startOfToken to: source position - 1).].
		token := token, (String with: self step)].

	token := token asSymbol! !

!LParser methodsFor: 'multi-character scans' stamp: 'spfa 11/18/2009 18:10'!
xColon		"process prolog :- and :? symbols"
	aheadChar = $- ifTrue:
		[self step.
		tokenType := #colonDash.
		self step.
		^ token := #colonDash].
	aheadChar = $? ifTrue:
		[self step.
		tokenType := #colonQuestion.
		self step.
		^ token := #colonQuestion].
	^super xColon! !

!LParser methodsFor: 'hack' stamp: 'spfa 12/21/2012 14:17'!
xLetter
	"Form a word or keyword."

	| type c |
	
	(SystemVersion current majorMinorVersion >= 'Squeak4.4')
		ifTrue: [^ super xLetter].
	
	buffer reset.
	[c := hereChar asciiValue.
	(type := typeTable at: c ifAbsent: [#xLetter]) == #xLetter or: [type == #xDigit]]
		whileTrue: ["open code step for speed"
			buffer nextPut: hereChar.
			hereChar := aheadChar.
			source atEnd
				ifTrue: [aheadChar := 30 asCharacter
					"doit"]
				ifFalse: [aheadChar := source next]].
	(type == #colon or: [type == #xColon and: [aheadChar ~= $=]])
		ifTrue: [buffer nextPut: self step.
			["Allow any number of embedded colons in literal symbols"
			(typeTable at: hereChar asciiValue ifAbsent: [#xLetter])
				== #xColon]
				whileTrue: [buffer nextPut: self step].
			tokenType := #keyword]
		ifFalse: [tokenType := #word].
	token := buffer contents.
	token isOctetString ifTrue: [token := token asOctetString].
! !
