SystemOrganization addCategory: #'Prolog-Core'!


Workspace subclass: #PrologWorkspace
	instanceVariableNames: 'doItReceiverPrototype'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Core'!

HierarchyBrowser subclass: #PrologBrowser
	instanceVariableNames: 'repl'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Core'!

PrologBrowser class
	instanceVariableNames: ''!

PrologWorkspace class
	instanceVariableNames: ''!

PackageInfo subclass: #PrologInfo
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Core'!

PrologInfo class
	instanceVariableNames: ''!

Logic subclass: #Prolog
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Core'!

Prolog subclass: #Family
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Core'!

Prolog subclass: #Doctor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Core'!

Prolog class
	instanceVariableNames: 'LocalDatabase Predicates BinaryPredicates'!

Doctor class
	instanceVariableNames: ''!

Family class
	instanceVariableNames: ''!

REPLMorph subclass: #PrologREPLMorph
	instanceVariableNames: 'maxAnswers'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Core'!

PrologREPLMorph class
	instanceVariableNames: ''!


!Doctor commentStamp: 'spfa 10/7/2022 18:27' prior: 0!
A demo class.

Used by Family>>health:
!
]style[(39)!

!Family commentStamp: 'spfa 10/7/2022 18:28' prior: 0!
A demo class

Examples:
LCompiler evaluate: 'Family new :? father(X,Y)'
LCompiler evaluate: 'Family new :? grandPa(''John'',Y)'
!
]style[(128)!

!Prolog commentStamp: 'spfa 10/8/2022 00:04' prior: 0!
This class and some of the related classes in the Prolog package were originally written by Mike Teng <miket@objectshare.com>, then at Digitalk, as part of the Prolog/V implementation distributed with Smalltalk V/286 in 1988.  We gratefully acknowledge Mike's permission to port his code to Squeak.

The port to Squeak is the work of Bolot Kerimbaev <bolot@cc.gatech.edu>, a summer intern at Walt Disney Imagineering. This was around 1999/2000.

From 2009 on, the code has then been maintained, refactored and extended by Stephane Rollandin <hepta@zogotounga.net>!
]style[(563)!

!PrologWorkspace commentStamp: 'spfa 2/26/2017 13:03' prior: 0!
PrologWorkspace

Enables "pure" Prolog syntax to be used, instead of directly invoking LCompiler

Try these:

Family new :? father(x,y)
Family new :? grandPa('John',y)
!
]style[(168)!

!Prolog class methodsFor: 'querying' stamp: 'spfa 9/28/2022 14:47'!
? anObject
	"Evaluating anObject in the context of an instance of this class."

	^ self new ? anObject! !

!Prolog methodsFor: 'evaluation' stamp: 'spfa 9/28/2022 14:58'!
? anObject
	"Syntactic sugar"

	^ self evaluate: anObject! !

!Prolog class methodsFor: 'prolog-compilation' stamp: 'bolot 7/12/1999 19:22'!
acceptsLoggingOfCompilation
	^true! !

!Prolog methodsFor: 'interaction' stamp: 'spfa 11/18/2009 18:15'!
addItemsToREPLMenu: aMenu.

	aMenu addLine; 
		add: 'explore database' target: self database selector: #explore;
		add: 'explore predicates cache' target: self class predicates selector: #explore;
		add: 'clear predicates cache' target: self class selector: #clearPredicates.
	! !

!PrologBrowser methodsFor: 'as yet unclassified' stamp: 'spfa 11/18/2009 18:15'!
addLowerPanesTo: window at: nominalFractions with: editString

	| verticalOffset row innerFractions m |

	row := AlignmentMorph newColumn
		hResizing: #spaceFill;
		vResizing: #spaceFill;
		layoutInset: 0;
		borderWidth: 1;
		borderColor: Color black;
		layoutPolicy: ProportionalLayout new.

	verticalOffset := 0.
	innerFractions := 0@0 corner: 1@0.
	verticalOffset := self addOptionalAnnotationsTo: row at: innerFractions plus: verticalOffset.
	verticalOffset := self addOptionalButtonsTo: row  at: innerFractions plus: verticalOffset.

	row 
		addMorph: ((m := self buildMorphicCodePaneWith: editString) borderWidth: 0)
		fullFrame: (
			LayoutFrame 
				fractions: (innerFractions withBottom: 1) 
				offsets: (0@verticalOffset corner: 0@0)
		).
	window 
		addMorph: row
		frame: nominalFractions.

	row on: #mouseEnter send: #paneTransition: to: window.
	row on: #mouseLeave send: #paneTransition: to: window.

	m hasUnacceptedEdits: false! !

!PrologInfo methodsFor: 'as yet unclassified' stamp: 'spfa 11/18/2009 18:15'!
additionalFiles

	^ #('PrologBrowser for ToolBuilder.cs')! !

!PrologInfo methodsFor: 'as yet unclassified' stamp: 'spfa 9/26/2022 21:27'!
additionalPackages
	"These are the names of packages that get loaded after the main change set"
	^#('Prolog-Documentation' 'Prolog-Dungeon' 'Prolog-Tests') ! !

!Prolog methodsFor: 'evaluation' stamp: 'spfa 9/28/2022 16:03'!
answer: aQuery

	"Return either a Boolean, or a list of dictionaries with the unification bindings.
	aQuery may either be a string, or a PrologList"

	^ self inferBindings: aQuery onBoolean: [:flag | ^ flag]! !

!Prolog methodsFor: 'arithmetic' stamp: 'spfa 10/14/2022 11:23'!
arithmeticFunctions

	^ #(+ - * / sqrt sin cos tan // \\ abs reciprocal sign negated)! !

!Prolog methodsFor: 'Smalltalk interop' stamp: 'spfa 10/14/2022 11:17'!
arithmeticValueOf: aCons

	aCons isNumber ifTrue: [^ aCons].

	(self arithmeticFunctions includes: aCons first)
		ifFalse: [self error: 'Unknow arithmetic function: ', aCons first].

	aCons length = 2 ifTrue: [
		^ (self arithmeticValueOf: aCons second logicValue)
			perform: aCons first].

	^ (self arithmeticValueOf: aCons second logicValue)
			perform: aCons first 
			with: (self arithmeticValueOf: aCons third logicValue).! !

!Prolog class methodsFor: 'database' stamp: 'spfa 11/18/2009 18:15'!
assert: anArray

	^ self assertz: anArray! !

!Prolog methodsFor: 'database' stamp: 'spfa 9/26/2022 09:11'!
assert: anArray

	^ self assertz: anArray! !

!Prolog class methodsFor: 'database' stamp: 'spfa 9/26/2022 09:10'!
asserta: anArray

	^ self database asserta: anArray! !

!Prolog methodsFor: 'database' stamp: 'spfa 9/26/2022 09:11'!
asserta: anArray

	^ self database asserta: anArray! !

!Prolog class methodsFor: 'database' stamp: 'spfa 9/26/2022 09:10'!
assertz: anArray

	^ self database assertz: anArray! !

!Prolog methodsFor: 'database' stamp: 'spfa 9/26/2022 09:11'!
assertz: anArray

	^ self database assertz: anArray! !

!Prolog class methodsFor: 'predicates' stamp: 'spfa 9/26/2022 17:06'!
binaryPredicates

	^ BinaryPredicates ifNil: [BinaryPredicates := self binaryPredicatesList as: Dictionary]! !

!Prolog class methodsFor: 'predicates' stamp: 'spfa 10/20/2022 01:07'!
binaryPredicatesList

	^  {
		 #= -> #unify.
		 #\+ -> #not.
		 #=** -> #univ.
		 #=== -> #idem.
		 #== -> #equal.
		 #\== -> #notEqual.
		 #>= -> #ge.
		 #> -> #gt.
		 #=< -> #le.
		 #< -> #lt.

		 #':-' -> #rule.
		 #',' -> #and
	} ! !

!PrologREPLMorph methodsFor: 'as yet unclassified' stamp: 'spfa 10/13/2022 17:13'!
buildMenu: aMenu

	super buildMenu: aMenu.
	
	aMenu addLine;
		add: 'max. answers: ', self maxAnswers action: #setMaxAnswers.

	^ aMenu! !

!PrologBrowser methodsFor: 'as yet unclassified' stamp: 'spfa 8/16/2009 23:02'!
classListIndex: anInteger 

	super classListIndex: anInteger .
	self selectedClass ifNil: [^self].
	(self selectedClass == Prolog
		or: [self selectedClass inheritsFrom: Prolog])
		ifTrue: [self repl evaluator: self selectedClass new]
	! !

!Prolog class methodsFor: 'predicates' stamp: 'spfa 9/29/2022 00:21'!
clearPredicates

	Predicates := BinaryPredicates := nil! !

!Prolog class methodsFor: 'prolog-compilation' stamp: 'spfa 9/26/2022 00:45'!
compilePredicate: aString

	self compile: (String streamContents: [:str |
		str nextPutAll: 'logic'; nextPutAll: (aString copyUpTo: $() capitalized; cr; cr.
		str nextPutAll: (aString copyFrom: $()])! !

!Prolog class methodsFor: 'prolog-compilation' stamp: 'bolot 7/7/1999 20:15'!
compilerClass
	^LCompiler! !

!PrologBrowser methodsFor: 'as yet unclassified' stamp: 'spfa 4/18/2011 13:54'!
createHierarchyTreeOf: col

	"used for Squeak version 4.3"
	^ super createHierarchyTreeOf: (col copyAfter: Prolog)! !

!Prolog class methodsFor: 'database' stamp: 'spfa 11/18/2009 18:15'!
database

	^ self databaseScope caseOf:
	{
		[#global] -> [Database].
		[#classLocal] -> [self localDatabase].
		[#instanceLocal] -> [self error: 'instance-local database (out of scope)'].
	} 
	otherwise: [self error: 'undefined database scope']! !

!Prolog methodsFor: 'database' stamp: 'spfa 11/18/2009 18:15'!
database

	^ self databaseScope caseOf:
	{
		[#global] -> [Database].
		[#classLocal] -> [self class localDatabase].
		[#instanceLocal] -> [self localDatabase].
	} 
	otherwise: [self error: 'undefined database scope']! !

!Prolog class methodsFor: 'database' stamp: 'spfa 11/18/2009 18:15'!
databaseScope

	^ #global! !

!Prolog methodsFor: 'database' stamp: 'spfa 11/18/2009 18:15'!
databaseScope

	^ self class databaseScope! !

!Prolog class methodsFor: 'prolog-compilation' stamp: 'spfa 11/18/2009 18:15'!
decompilerClass
	"Answer a decompiler class appropriate for compiled methods of this class."

	^LDecompiler! !

!PrologWorkspace class methodsFor: 'as yet unclassified' stamp: 'spfa 10/4/2022 22:26'!
demoString
	^
'Family example:
------------------------

Family new :? father(X,Y)
Family new :? grandPa(''John'',Y)

Family new :? father(X, ''Mary'')
Family new :? father(''John'', ''Mary'')

Family new :? grandPa(''John'', X), father(Y, X)

Family new :? health(''John'', X).
Family new :? consult( healthHistory(''John'', X), Doctor new)


GUI:
-----

Family new openGui


Database:
---------------

Family new :? retractall(brother).

Family new :? assertz(brother(''John'', ''Stewart'')), asserta(brother(''John'', ''Barbara'')), assertz(brother(''John'', ''Alan'')).

Family new :? brother(X,Y). 

Family new :? retract(brother(''John'', ''Stewart'')).

Family new :? brother(X,Y).


Dungeon
--------------

Dungeon new :? go


Resources:
----------------

Prolog Tutorial
https://www.cpp.edu/~jrfisher/www/prolog_tutorial/contents.html
'! !

!PrologREPLMorph methodsFor: 'as yet unclassified' stamp: 'spfa 9/29/2022 15:23'!
doCommand: aString

"TO DO: see if this version should be the super implementation"

	| answer conlen nextContents |

	self setProperty: #inCommand toValue: true.
	self registerCommand: aString.

	conlen := myContents size.

	debugContinuation := lastAnswer := lastError := lastValue :=  nil.

	answer :=  [lastAnswer := (self formatAnswer: 
						(lastValue := self evalString: aString)) 
								asText addAttributes: self class answerAttributes]
		on: Error
		do: [:ex | 
				debugContinuation := Continuation current.
				debugContinuation isContinuation
					ifTrue: [(lastError := self exceptionDescription: ex)
								asText addAttribute: self class errorAttribute]
					ifFalse: [ex signal]].

	self setProperty: #inCommand toValue: false.

	self appendEntry; outputQueue.
	nextContents := myContents.
	nextContents size = conlen ifTrue: [nextContents := nextContents allButLast].
	self updateText: nextContents, self separator, String cr, answer, String cr.
	self refreshDisplay: self prompt.
! !

!PrologWorkspace methodsFor: 'as yet unclassified' stamp: 'bolot 12/13/2002 15:46'!
doItReceiver
	^doItReceiverPrototype ifNil: [doItReceiverPrototype := Prolog new]! !

!PrologWorkspace methodsFor: 'as yet unclassified' stamp: 'spfa 11/18/2009 18:15'!
doItReceiver: aProlog
	doItReceiverPrototype := aProlog! !

!Prolog methodsFor: 'input-output' stamp: 'spfa 11/18/2009 18:15'!
doTranscript

	self outputStream: Transcript! !

!Prolog methodsFor: 'interaction' stamp: 'spfa 10/17/2022 11:57'!
evalString: aString notifying: requestor

"TO DO: pass the requestor to the compilation cue"

	| gen results |

	gen := self generatorOnQuery: aString withBindings: true.

	results := (gen next: requestor maxAnswers) select: #notNil.

	gen atEnd ifFalse: [^ results, {{#'... and more ...'}}].

	^ results ! !

!Prolog class methodsFor: 'querying' stamp: 'spfa 9/28/2022 14:48'!
evaluate: anObject
	"Evaluating anObject in the context of an instance of this class."

	^ self new evaluate: anObject! !

!Prolog methodsFor: 'evaluation' stamp: 'spfa 9/28/2022 15:32'!
evaluate: anObject
	"Evaluating anObject in the context of this instance."

	^ self evaluate: anObject withBindings: false! !

!Prolog class methodsFor: 'querying' stamp: 'spfa 9/28/2022 15:33'!
evaluate: anObject getBindings: aString
	"Evaluating anObject in the context of an instance of this class."

	^ self new evaluate: anObject getBindings: aString! !

!Prolog methodsFor: 'evaluation' stamp: 'spfa 9/28/2022 15:32'!
evaluate: anObject getBindings: aString

	| results |

	results := self evaluate: anObject withBindings: true.

	results ifNil: [^ false].
	results ifEmpty: [self error: 'No unification for ', aString].

	^ Array streamContents: [:str |
		results do: [:result |
			result do: [:ea | 
				(ea class == LogicResult and: [ea term = aString])
					ifTrue: [str nextPut: ea logicResult]]]].
! !

!Prolog class methodsFor: 'querying' stamp: 'spfa 9/28/2022 15:36'!
evaluate: anObject withBindings: aBoolean
	"Evaluating anObject in the context of an instance of this class."

	^ self new evaluate: anObject withBindings: aBoolean! !

!Prolog methodsFor: 'evaluation' stamp: 'spfa 10/13/2022 16:04'!
evaluate: anObject withBindings: aBoolean
	"Evaluating anObject in the context of this instance."

	| results |

	results := (self generatorOnQuery: anObject withBindings: aBoolean) upToEnd.
	results ifEmpty: [^ nil].

	^ results asArray! !

!Prolog class methodsFor: 'prolog-compilation' stamp: 'bolot 8/5/1999 16:21'!
evaluatorClass
	"Answer an evaluator class appropriate for evaluating expressions in the 
	context of this class."

	^LCompiler! !

!PrologREPLMorph methodsFor: 'as yet unclassified' stamp: 'spfa 9/26/2022 11:44'!
exceptionDescription: anException

	^ anException description! !

!Prolog methodsFor: 'input-output' stamp: 'spfa 11/18/2009 18:15'!
finalizeOutput

	outputStream == Transcript
		ifTrue: [outputStream endEntry]! !

!PrologBrowser class methodsFor: 'as yet unclassified' stamp: 'spfa 8/17/2009 13:57'!
for: aProlog
	"Return a new PrologBrowser on the given class"
	|  newBrowser window |
	newBrowser := PrologBrowser new initHierarchyForClass: Prolog.
	newBrowser selectClass: aProlog class..
	newBrowser repl: (PrologREPLMorph on: aProlog).
	newBrowser editSelection: #editComment.
	window _ MorphicToolBuilder new build: (newBrowser openSystemCatEditString: nil).
	window setLabel: newBrowser labelString, ' & REPL' .
	^ window
! !

!PrologREPLMorph methodsFor: 'as yet unclassified' stamp: 'spfa 10/2/2022 23:34'!
formatAnswer: anObject

	anObject isEmptyOrNil ifTrue: [^ 'No.']. "can be empty after a cut"
	anObject = #(()) ifTrue: [^ 'Yes.'].

	^ String streamContents: [:str |
		anObject do: [:elt |
			elt size isZero ifTrue: [str nextPutAll: 'Yes.']
				ifFalse: [
					elt size = 1 ifTrue: [elt first printForPrologOn: str]
						ifFalse: [elt printElementsForPrologOn: str]
				].
			str cr].
		str skip: -1]
! !

!Prolog methodsFor: 'evaluation' stamp: 'spfa 10/13/2022 15:54'!
generatorOnQuery: anObject withBindings: aBoolean
	"Evaluating anObject in the context of this instance."

	anObject isString ifTrue: [
		^ self class compilerClass generatorOnQueryTo: self for: anObject withBindings: aBoolean].

	^ Generator on: [:g |
		[self inferSmalltalk: (PrologList car: anObject asPrologRelation)]
			on: PrologAnswerStep 
			do: [:ex | 
				g yield: (aBoolean ifTrue: [ex answer] 
							ifFalse: [self class compilerClass deResultify: ex answer]).
				ex resume]]! !

!Prolog methodsFor: 'as yet unclassified' stamp: 'spfa 10/11/2022 22:40'!
getGoal: anObject

	"Here we silently promote a Symbol into a goal (a Relation), which we can afford to do because a goal is precisely what we always expect to find (it is the responsibility of the sender to not be wrong about this). 	
	This makes it possible to use /0 predicates without any parenthesis"

	anObject isSymbol ifTrue: [^ Relation car: anObject].

	anObject isPrologList 
		ifFalse: [self error: 'expected a callable atom'].

	^ anObject! !

!Prolog methodsFor: 'interaction' stamp: 'spfa 8/17/2009 13:53'!
gui

	^ PrologBrowser for: self
! !

!Prolog methodsFor: 'interaction' stamp: 'spfa 11/18/2009 18:15'!
guiColor

	^ Color blue muchLighter! !

!Prolog methodsFor: 'interaction' stamp: 'spfa 11/18/2009 18:15'!
guiDocumentation

	^'"
Use the read-eval-print loop listener below to query the currently selected Prolog subclass.
Choose the ''decompile'' display option to inspect the actual Smalltalk code for logic predicates
(that''s the right-most button)

"'! !

!Prolog methodsFor: 'evaluation' stamp: 'spfa 9/28/2022 16:41'!
inferBindings: aQuery into: bBlock onBoolean: aBlock

	"Interpret the query results as either Booleans, in which case they are processed by aBlock, or as a series of OrderedDictionaries collecting the (then twice-ordered) unification bindings, in which case each dictionary is in turn processed by bBlock.

	To make it clear about ordering: 
	- the dictionaries are ordered just as the answers were found
	- each dictionary has its keys in their order of appearance, from left to right, in the query"

	(self evaluate: aQuery withBindings: true) in: [:res |
		res ifNil: [^ aBlock value: false].
		(res size = 1 and: [res anyOne isEmpty]) ifTrue: [^ aBlock value: true].
		res do: [:ea | bBlock value: (ea as: OrderedDictionary)]]! !

!Prolog methodsFor: 'evaluation' stamp: 'spfa 9/28/2022 15:51'!
inferBindings: anObject onBoolean: aBlock

	^ Array streamContents: [:str |
		self inferBindings: anObject into: [:res | str nextPut: res] onBoolean: [:flag | ^ aBlock value: flag]]! !

!PrologBrowser methodsFor: 'as yet unclassified' stamp: 'spfa 4/18/2011 13:48'!
initHierarchyForClass: aClassOrMetaClass
	| tab stab index nonMetaClass |
	
	"hack for compatibility up to version 4.3"
	SystemVersion current highestUpdate >= 11314
		 ifTrue: [^ super initHierarchyForClass: aClassOrMetaClass].
	
	centralClass := aClassOrMetaClass.
	nonMetaClass := aClassOrMetaClass theNonMetaClass.
	self systemOrganizer: SystemOrganization.
	metaClassIndicated := aClassOrMetaClass isMeta.
	classList := OrderedCollection new.
	tab := ''.
	nonMetaClass allSuperclasses reverseDo: 
		[:aClass | 
		(Prolog inheritsFrom: aClass)
			ifFalse: [
				classList add: tab , aClass name.
				tab := tab , '  ']].
	index := classList size + 1.
	nonMetaClass allSubclassesWithLevelDo:
		[:aClass :level |
		stab := ''.  1 to: level do: [:i | stab := stab , '  '].
		classList add: tab , stab , aClass name]
	 	startingLevel: 0.
	self classListIndex: index! !

!Prolog methodsFor: 'interaction' stamp: 'spfa 9/26/2022 11:03'!
labelForREPL

	^ self name ! !

!PrologBrowser methodsFor: 'as yet unclassified' stamp: 'spfa 11/18/2009 18:15'!
labelString

	^ 'Prolog Browser'! !

!Prolog class methodsFor: 'database' stamp: 'spfa 9/26/2022 09:34'!
localDatabase

	^ LocalDatabase ifNil: [LocalDatabase := PrologDatabase new]! !

!Prolog methodsFor: 'database' stamp: 'spfa 9/26/2022 09:15'!
localDatabase

	^ database ifNil: [database := PrologDatabase new]! !

!Prolog methodsFor: 'basics' stamp: 'spfa 10/13/2022 00:26'!
logicAppend

append([], X, X). 
append([H | X], Y, [H | Z]) :- append(X, Y, Z).! !

!Prolog methodsFor: 'structures' stamp: 'spfa 10/11/2022 15:48'!
logicArg

"hack see #testFunctor2"
arg(N, S, X) :-
	integer(N),
	iss(T, (s logicValue nth: n) class == LogicDontCare),
	test(T),
	iss(V, (s logicValue nth: n) becomeForward: LogicRef new),
	=(X,V),!!.

arg(N, S, X) :-
	integer(N),
	iss(T, (s logicValue nth: n) class == LogicRef),
	test(T),
	iss(V, (s logicValue nth: n)),
	=(X,V),!!.

arg(N, Structure, X) :-
	iss(List, structure logicInstanciated),
	iss(Len, list logicValue length - 1),
	numlist(0,Len,Range),
	member(N, Range),
	iss(X, (list logicValue nth: n logicValue)), !!.

arg(N, Structure, X) :-
	integer(N),
	integer(X),
	iss(X, structure nth: n).

! !

!Prolog methodsFor: 'classifying terms' stamp: 'spfa 10/11/2022 21:54'!
logicAtomic

atomic(Term) :-
	nonvar(Term),
	\+ compound(Term).! !

!Prolog methodsFor: 'basics' stamp: 'spfa 10/14/2022 10:52'!
logicBetween

between(N, inf, K) :- K = N.
between(N, inf, K) :- N1 is N+1, between(N1, inf, K).

between(N, infinity, K) :- between(N, inf, K).

between(N, M, K) :- N =< M, K = N.
between(N, M, K) :- N < M, N1 is N+1, between(N1, M, K).
! !

!Prolog methodsFor: 'definite clause grammar' stamp: 'spfa 10/6/2022 21:56'!
logicCccc

cccc([X|S], X, S).! !

!Family methodsFor: 'demo' stamp: 'spfa 11/18/2009 18:09'!
logicFather

father('John', 'Mary').  "John is the father of Mary"
father('John', 'David').
father('David', 'Jack').
father('Arthur', 'Nancy').
! !

!Prolog methodsFor: 'structures' stamp: 'spfa 10/13/2022 00:26'!
logicFunctor
	"The first argument can be either a
	Relation or a List whose first
	element is a Symbol."

functor(Structure, Head, Arguments) :-
	univ(Structure, [H | T]),
	atomic(H),
	=(Head, H),
	iss(Arguments, t logicValue ifNil: [0] ifNotNil: #length).

functor(Structure, Head, Arguments) :-
	var(Structure),
	atomic(Head),
	integer(Arguments),
	iss(Structure, self
		structureHead: head logicValue
		arguments: arguments logicValue).
! !

!Family methodsFor: 'demo' stamp: 'spfa 10/4/2022 00:27'!
logicGrandPa

"X is the grandPa of Y."

grandPa(X, Y) :-
	father(X, Z),
	or(father(Z, Y),mother(Z,Y)).! !

!Family methodsFor: 'demo' stamp: 'spfa 10/4/2022 00:27'!
logicHealth

"Return the health history of X."

health(X, History) :-
	consult(healthHistory(X, History), Doctor new).! !

!Doctor methodsFor: 'as yet unclassified' stamp: 'spfa 11/18/2009 18:09'!
logicHealthHistory

healthHistory('John', 'John had two times of heart attacks').
healthHistory('Mary', 'Mary had no serious problems').! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 10/13/2022 14:54'!
logicLast

last(X,Y) :-
    append(_,[Y],X).! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 10/13/2022 10:09'!
logicListLength

listLength([], 0).
listLength([_|Ls], N) :-
	integer(N),
       >(N, 0),
       iss(N0, n - 1),
       listLength(Ls, N0).


listLength([H|T], N) :-
	var(N),
	=(T,[]).

listLength([H|T], N) :-
	var(N),
	=(T,[])",
	length(H,N)".! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 10/13/2022 18:32'!
logicListLength2

listLength2(Xs,L) :- listLength2(Xs,0,L) .

listLength2( []     , L , L ) .
listLength2( [_|Xs] , T , L ) :-
  T1 is T+1 ,
  listLength2(Xs,T1,L)
  .! !

!Prolog methodsFor: 'basics' stamp: 'spfa 10/3/2022 22:03'!
logicMember

member(X,[X|:=]).
member(X,[:=|R]) :- member(X,R).! !

!Family methodsFor: 'demo' stamp: 'spfa 10/4/2022 00:27'!
logicMother

"Mary is the mother of Nancy."

mother('Mary', 'Nancy').! !

!Family methodsFor: 'demo' stamp: 'spfa 10/4/2022 00:28'!
logicParent

parent(X,Y) :-
	or(father(X,Y), mother(X,Y)).! !

!Prolog methodsFor: 'definite clause grammar' stamp: 'spfa 10/7/2022 17:41'!
logicPhrase


phrase(DCG,List) :- atom(DCG),
					append([DCG],[List,S],FullUL),
					univ(Callable,FullUL),
					call(Callable),
					=(S,[]),!!.

phrase(DCG,List) :- not(atom(DCG)),
					univ(DCG,UL),
					append(UL,[[_|List],[]],FullUL),
					univ(Callable,FullUL),
					call(Callable),!!.
! !

!Prolog methodsFor: 'structures' stamp: 'spfa 10/13/2022 00:26'!
logicUniv


univ(Structure, List) :-
	nonvar(Structure),
	send(Structure, asPrologRelation, S),
	iss(List, PrologList
		head: s head
		tail: s tail),
	!!.


univ(Structure, [Head | List]) :-
	atom(Head),
	iss(X, list),
	iss(Structure, Relation
		head: head logicValue
		tail: x logicValue),!!.



univ(List, Term) :-
	var(List),
	iss(X, term),
	iss(List, PrologList
		head: x logicValue head
		tail: x logicValue tail).

! !

!PrologREPLMorph methodsFor: 'as yet unclassified' stamp: 'spfa 10/13/2022 17:11'!
maxAnswers

	^ maxAnswers ifNil: [10]
! !

!PrologREPLMorph methodsFor: 'as yet unclassified' stamp: 'spfa 10/13/2022 17:15'!
maxAnswers: aNumber

	maxAnswers := (aNumber rounded max: 1)! !

!PrologBrowser class methodsFor: 'as yet unclassified' stamp: 'spfa 8/17/2009 13:59'!
newFor: aPrologClass
	"Open a new PrologBrowser on the given class"
	|  newBrowser |
	newBrowser := PrologBrowser new initHierarchyForClass: Prolog.
	newBrowser selectClass: aPrologClass..
	newBrowser repl: (PrologREPLMorph on: aPrologClass new).
	newBrowser editSelection: #editComment.
	Browser openBrowserView: (newBrowser openSystemCatEditString: nil)
		label: newBrowser labelString, ' & REPL' 
! !

!Prolog class methodsFor: 'prolog-compilation' stamp: 'spfa 11/18/2009 18:15'!
noteCompilationOf: aSelector meta: isMeta
	
	self withAllSubclassesDo: [:pl | pl clearPredicates]! !

!PrologREPLMorph methodsFor: 'as yet unclassified' stamp: 'spfa 11/18/2009 18:15'!
notify: string at: location in: source

	"in you see this in a debugger, printIt for details:"
	source contents copyReplaceFrom: location to: location - 1
							with: string. "printIt"

	^ self error: 'bad syntax'

! !

!PrologWorkspace class methodsFor: 'as yet unclassified' stamp: 'bolot 12/13/2002 15:45'!
open
	self openLabel: 'Prolog Workspace'! !

!PrologWorkspace class methodsFor: 'as yet unclassified' stamp: 'spfa 11/18/2009 18:15'!
openDemo
	"self openDemo"
	(self new contents: self demoString) openLabel: 'Prolog Workspace'! !

!Prolog methodsFor: 'interaction' stamp: 'spfa 11/18/2009 18:15'!
openGui

	self gui openInWorld.! !

!Prolog methodsFor: 'interaction' stamp: 'spfa 9/26/2022 11:03'!
openREPL

	((PrologREPLMorph on: self) inSystemWindowLabelled: self labelForREPL 
		color: "Color perrywinkle muchLighter "(Color r: 0.897 g: 0.895 b: 0.978)) 
			openInWorld! !

!PrologBrowser methodsFor: 'as yet unclassified' stamp: 'spfa 8/17/2009 00:08'!
openSystemCatEditString: aString
	"Create a pluggable version of all the views for a Browser, including views and controllers.  The top list view is of the currently selected system class category--a single item list."
	"Example:
		Browser new browseAllClasses.
	"
	| builder max |
	builder := ToolBuilder default.
	max := self wantsOptionalButtons ifTrue:[0.32] ifFalse:[0.4].
	^self buildWindowWith: builder specs: {
		(0.0@0.0 corner: 0.333@(max-0.1)) -> [self buildClassListWith: builder].
		(0.0@(max-0.1) corner: 0.333@max) -> [self buildSwitchesWith: builder].
		(0.333@0.0 corner: 0.666@max) -> [self buildMessageCategoryListWith: builder].
		(0.666@0.0 corner: 1@max) -> [self buildMessageListWith: builder].
		(0@max corner: 1@0.8) -> [self buildCodePaneWith: builder].
		(0@0.8 corner: 1@1) -> [PluggableMorphSpec new morph: self repl].
	}! !

!PrologREPLMorph methodsFor: 'as yet unclassified' stamp: 'spfa 9/29/2022 15:27'!
pointedCommand

	^ (super pointedCommand ifNil: [^ nil]) lines first! !

!PrologInfo methodsFor: 'as yet unclassified' stamp: 'spfa 11/18/2009 18:15'!
postscriptText

	^'
	"check if PrologBrowser needs to use ToolBuilder in the target image"
	(HierarchyBrowser includesSelector: #buildWith:) ifTrue:
		[FileStream fileIn: ''PrologBrowser for ToolBuilder.cs''].
	'! !

!Prolog class methodsFor: 'predicates' stamp: 'spfa 11/18/2009 18:15'!
predicates

	^ Predicates ifNil: [Predicates := IdentityDictionary new]! !

!PrologInfo methodsFor: 'as yet unclassified' stamp: 'spfa 11/30/2011 14:58'!
prerequisitePackages
	"These are the names of packages that get loaded before the main change set"
	^#('ConsCell' 'Prolog-Compatibility' 'Prolog-Compiler' 'Continuation' 'REPLMorph' 'Prolog-Base' 'Prolog-Tools') ! !

!Prolog methodsFor: 'database' stamp: 'spfa 10/3/2022 15:00'!
primlogicAbolish: assoc
        "Purge a database identified by the
         structure in the argument."
	| arg |
	self check: assoc forArity: 1.
	arg := self first: assoc.
	arg isSymbol ifTrue: [arg := Relation car: arg].
	((arg isRelation) and: [arg head isSymbol])
		ifFalse: [self error: 'argument is not a fact'].
	self database removeKey: arg head.
	^assoc value! !

!Prolog methodsFor: 'complex goals' stamp: 'spfa 11/18/2009 18:15'!
primlogicAnd: assoc
    "Try to satisfy a list of goals."
    | aList next |
    aList := assoc key.
	aList ifNil: [^ assoc value].
	next := (Relation with: #primlogicAnd:) tail: aList tail.
   	self doGoal: aList head
          	continue: [self doGoal: next continue: assoc value]

 ! !

!Prolog methodsFor: 'database' stamp: 'spfa 11/18/2009 18:15'!
primlogicAssert: assoc
        "Add the fact in assoc to the
         end of the database."
    ^ self primlogicAssertz: assoc! !

!Prolog methodsFor: 'database' stamp: 'spfa 9/28/2022 22:12'!
primlogicAsserta: assoc
        "Add the fact in assoc to the
         beginning of the database."
    | arg aList |
    self check: assoc forArity: 1.
    arg := (self first: assoc) logicInstanciated.
    aList := self database at: arg head ifAbsent:  [
		self database at: arg head put: (PrologList car: arg tail).
		^ assoc value].
    self database at: arg head put:
        (PrologList head: arg tail tail: aList).
    ^assoc value! !

!Prolog methodsFor: 'database' stamp: 'spfa 9/28/2022 22:13'!
primlogicAssertz: assoc
        "Add the fact in assoc to the
         end of the database."
    | arg aList |
    self check: assoc forArity: 1.
    arg := (self first: assoc) logicInstanciated.
    aList := self database at: arg head ifAbsent: [
		self database at: arg head put: (PrologList car: arg tail).
		^ assoc value].
    aList ifNil: [self database at: arg head put: (PrologList car: arg tail)] 
		ifNotNil: [aList addLast: arg tail].
    ^assoc value! !

!Prolog methodsFor: 'classifying terms' stamp: 'spfa 11/18/2009 18:15'!
primlogicAtom: assoc
        "Succeed if the argument is a Symbol."
    self check: assoc forArity: 1.

    (self first: assoc) isSymbol
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'higher-order predicates' stamp: 'spfa 10/13/2022 16:31'!
primlogicBagof: assoc

	"No support for existential quantifiers at the moment"

    	| goal answers template var bag mark cachedVarList |

   	self check: assoc forArity: 3.

	template := assoc key head.
	template isPrologList ifFalse: [template := nil].
	bag := OrderedCollection new.

	goal := self getGoal: (self second: assoc).
	var := assoc key tail tail head .

	cachedVarList := varList copy.
	self addVarFrom: assoc key. 
	mark := unboundVars size.

	answers := self doGoal: goal onSuccess: [
		var getValueFrom: {bag add: (template 
									ifNotNil: [template resultCopy] 
									ifNil: [assoc key head resultCopy])} asPrologList.
		self inferAnswer: false].
	[answers isClosure] whileTrue: [answers := answers value].

	varList := cachedVarList.
	self unbindVar: mark. 

	bag ifEmpty: [^ self "fails"].

	template ifNil: ["works just like findall/3"
			var getValueFrom: bag asPrologList.
			answer := OrderedCollection new]
		ifNotNil: [answer removeLast].	

	^ assoc value! !

!Prolog methodsFor: 'higher-order predicates' stamp: 'spfa 10/3/2022 22:55'!
primlogicCall: assoc
        "Execute the argument as a predicate."
    | goal |
    self check: assoc forArity: 1.
    goal := self getGoal: (self first: assoc).
    ^[self doGoal: goal onSuccess: assoc value]! !

!Prolog methodsFor: 'classifying terms' stamp: 'spfa 10/11/2022 21:58'!
primlogicCompound: assoc

	self check: assoc forArity: 1.
	(self first: assoc) isPrologList ifTrue: [^assoc value]! !

!Prolog methodsFor: 'Smalltalk interop' stamp: 'spfa 10/3/2022 22:56'!
primlogicConsult: assoc
      "Ask the second argument (a Smalltalk expression) 
	to solve the goals contained in the first argument, e.g.
            consult(father(x, 'Mary'), Family new)."
	| goal |
	self check: assoc forArity: 2.
	goal := self getGoal: (self first: assoc).
	^ [(self second: assoc) value
		doGoal: goal onSuccess: assoc value]! !

!Prolog methodsFor: 'backtracking' stamp: 'spfa 9/29/2022 15:01'!
primlogicCut: assoc 
	" Succeed always, exit the containing method
	  during backtracking."
	| blk mark |
    	self check: assoc forArity: 1.
	self execute: assoc value.
	assoc key head value! !

!Prolog methodsFor: 'database' stamp: 'spfa 11/18/2009 18:15'!
primlogicDatabase: assoc
        "Create a database which can hold facts
         being asserted or extracted. The argument
         must be a structure."
    | arg |
    self check: assoc forArity: 1.
    arg := self first: assoc.
    (arg isConsCell and: [arg isRelation] and: [arg head isSymbol])
        ifFalse: [self error: 'argument is not a fact'].
    self database at: arg head ifAbsent: [
        self database at: arg head put: nil].
    ^assoc value! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 9/26/2022 15:20'!
primlogicEqual: assoc
        "Succeed if the first argument is equal to the second."
    self check: assoc forArity: 2.
    (self first: assoc) = (self second: assoc)
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'backtracking' stamp: 'spfa 9/29/2022 12:21'!
primlogicExit: assoc
	" exit Prolog "
	self execute: assoc value.
	^returnBlock value! !

!Prolog methodsFor: 'success-failure' stamp: 'spfa 11/18/2009 18:15'!
primlogicFail: assoc
        "Fail always."! !

!Prolog methodsFor: 'success-failure' stamp: 'spfa 12/1/2011 14:24'!
primlogicFalse: assoc
        "Fail always."! !

!Prolog methodsFor: 'higher-order predicates' stamp: 'spfa 10/13/2022 16:24'!
primlogicFindall: assoc

    	| goal answers template bag mark cachedVarList |

   	self check: assoc forArity: 3.

	template := assoc key head.
	template isPrologList ifFalse: [template := nil].
	bag := OrderedCollection new.

	goal := self getGoal: (self second: assoc).

	cachedVarList := varList copy.
	self addVarFrom: assoc key. "NEEDED (hack?) to get logicC to work"

	mark := unboundVars size.

	answers := self doGoal: goal 
					onSuccess: [

						self inferAnswer: false.
						answer first ifNotEmpty: [
							bag add: (template 
									ifNotNil: [
										template resultCopy "will not work with morphs!!"] 
									ifNil: [answer last first logicResult])]].

	[answers isClosure] whileTrue: [answers := answers value].

	assoc key tail tail head getValueFrom: bag asPrologList.

	varList := cachedVarList.
	self unbindVar: mark. 
	answer := OrderedCollection new.

	^ assoc value! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 11/18/2009 18:15'!
primlogicGe: assoc
        "Succeed if the first argument is
         greater than or equal to the second."
    self check: assoc forArity: 2.
    (self first: assoc) >= (self second: assoc)
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'classifying terms' stamp: 'spfa 9/29/2022 20:28'!
primlogicGround: assoc
        "Succeed if the argument has no unbound variables"
    self check: assoc forArity: 1.
    (self first: assoc) isLogicGround ifTrue: [^assoc value]! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 11/18/2009 18:15'!
primlogicGt: assoc
        "Succeed if the first argument is
         greater than the second."
    self check: assoc forArity: 2.
    (self first: assoc) > (self second: assoc)
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'debugging' stamp: 'spfa 11/18/2009 18:15'!
primlogicHalt: assoc
        "Halt the execution.  Smalltalk debugger
         can be invoked at this point."
    assoc key ifNotNil: [^self].
    self halt! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 9/29/2022 22:31'!
primlogicIdem: assoc
        "Succeed if the two arguments are the same object."
    self check: assoc forArity: 2.
    (self first: assoc) == (self second: assoc)
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'classifying terms' stamp: 'spfa 11/18/2009 18:15'!
primlogicInteger: assoc
        "Succeed if the argument is an Integer."
    self check: assoc forArity: 1.
    ((self first: assoc) isKindOf: Integer)
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'Smalltalk interop' stamp: 'spfa 10/11/2022 10:59'!
primlogicIs: assoc

	"Succeeds when the first argument is the value of the arithmetic expression represented by the second argument"

	self check: assoc forArity: 2.

	self unify: assoc key head 
		with: (self arithmeticValueOf: assoc key tail head) 
		then: [^assoc value]! !

!Prolog methodsFor: 'Smalltalk interop' stamp: 'spfa 10/11/2022 22:35'!
primlogicIss: assoc
        "Execute the second argument as a Smalltalk
         expression and then unify its result with
         the first argument."

	"Beware: Smalltalk precedence rules for arithmetic will apply. 
	See primlogicIs: for the standard is/2 implementation"

    self check: assoc forArity: 2.

    self unify: assoc key head 
        with: assoc key tail head value
        then: [^assoc value]! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 11/18/2009 18:15'!
primlogicLe: assoc
        "Succeed if the first argument is
         less than or equal to the second."
    self check: assoc forArity: 2.
    (self first: assoc) <= (self second: assoc)
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 10/13/2022 10:02'!
primlogicLength: assoc
      
"Bad implementation
see https://www.swi-prolog.org/pldoc/man?predicate=length/2
and #testLength
"

	| list |

	self check: assoc forArity: 2.

	list := assoc key head logicValue.

	list ifNil: [
    		self unify: 0
        		with: assoc key tail head
        		then: [^assoc value].
		^ self].

	(list isPrologList and: [list cdr isNil]) ifTrue: [
    		self unify: list length
        		with: assoc key tail head
        		then: [^assoc value].
		^ self].

   ^ self perform: #logicListLength with: assoc! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 11/18/2009 18:15'!
primlogicLt: assoc
        "Succeed if the first argument is
         less than the second."
    self check: assoc forArity: 2.
    (self first: assoc) < (self second: assoc)
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 11/18/2009 18:15'!
primlogicNe: assoc
        "Succeed if the first argument is
         greater than the second."
    self check: assoc forArity: 2..
    (self first: assoc) ~= (self second: assoc)
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'input-output' stamp: 'spfa 11/18/2009 18:15'!
primlogicNl: assoc
        "Write an end of line to the output stream"
    assoc key ifNotNil: [^self].
    self outputStream cr.
    self finalizeOutput.
    ^assoc value! !

!Prolog methodsFor: 'classifying terms' stamp: 'spfa 9/26/2022 14:08'!
primlogicNonvar: assoc
        "Succeed if the argument is not an
         unbound variable."
    self check: assoc forArity: 1.
   ((self first: assoc) class == LogicUnbound
		or: [(self first: assoc) class == LogicDontCare])
        ifFalse: [^assoc value]! !

!Prolog methodsFor: 'complex goals' stamp: 'spfa 10/3/2022 22:53'!
primlogicNot: assoc
        "Succeed if the evaluation of the argument
         as a goal fails."
	| goal mark |
     	self check: assoc forArity: 1.
    	goal := self getGoal: (self first: assoc).
	mark := unboundVars size.
	self execute: (self doGoal: goal onSuccess: [self unbindVar: mark. ^self]).
    	^assoc value! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 9/26/2022 15:20'!
primlogicNotEqual: assoc
        "Succeed if the first argument is not equal to the second."
    self check: assoc forArity: 2.
    (self first: assoc) = (self second: assoc)
        ifFalse: [^assoc value]! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 11/18/2009 18:15'!
primlogicNotEquiv: assoc
        "Succeed if the two arguments
         are not the same object."
    self check: assoc forArity: 2.
    (self first: assoc) == (self second: assoc)
        ifFalse: [^assoc value]! !

!Prolog methodsFor: 'basics' stamp: 'spfa 11/18/2009 18:15'!
primlogicNumlist: assoc
        "Unify the third argument with a list of integers 
		within a range defined by the first arguments."
    self check: assoc forArity: 3.
    self unify: assoc key tail tail head 
		with: ((self first: assoc) to: (self second: assoc)) asArray asPrologList 
        	then: [^assoc value]

! !

!Prolog methodsFor: 'complex goals' stamp: 'spfa 10/3/2022 16:26'!
primlogicOr: assoc
    "Disjunction predicate."
	| satisfied |
    assoc key length isZero ifTrue: [^self].
	satisfied := false.
	0 to: assoc key length - 1 do: [:i |	
		satisfied ifFalse: [
    			self doGoal: (assoc key nth: i) continue: [satisfied := true. assoc value value]]]! !

!Prolog methodsFor: 'input-output' stamp: 'spfa 11/18/2009 18:15'!
primlogicRead: assoc
        "Read a Smalltalk expression, evaluate it and
         answer the result.  This can be used to input
         a Symbol, a number, a quoted string, etc."
    | anAnswer |
    self check: assoc forArity: 1.
    anAnswer := Compiler evaluate: (FillInTheBlank
		request: 'type a Smalltalk expression and accept'
		initialAnswer: '').
    self unify: assoc key head
        with: anAnswer
        then: [^assoc value]! !

!Prolog methodsFor: 'input-output' stamp: 'spfa 11/18/2009 18:15'!
primlogicReadstring: assoc
        "Read a string and answer the result."
    | anAnswer |
    self check: assoc forArity: 1.

	[self inputStream atEnd] whileTrue: [World doOneCycle. Processor yield].
"
    anAnswer := FillInTheBlank
		request: 'type a line and accept'
		initialAnswer: ''."
	anAnswer := self inputStream contents.
 	self inputStream resetContents.

	self unify: assoc key head
        with: anAnswer
        then: [^assoc value]! !

!Prolog methodsFor: 'complex goals' stamp: 'spfa 9/29/2022 13:31'!
primlogicRepeat: assoc
        "Succeed always, even
         during backtacking."

    self check: assoc forArity: 0.
    [self execute: assoc value] repeat! !

!Prolog methodsFor: 'database' stamp: 'spfa 10/3/2022 14:52'!
primlogicRetract: assoc
        "Delete the first fact unified with the
         one in assoc from the database. Delete
         the next one during backtrack."
	| arg aList bList |

	self check: assoc forArity: 1.
	arg := self first: assoc.
	aList := self database at: arg head ifAbsent: [^self].
	aList ifNil: [^self].

	self unify: arg tail
		with: aList head
		then: [
			self database at: arg head put: aList tail.
			assoc value value].

	[(bList := aList tail) isNil] whileFalse: [
	self unify: arg tail
		with: bList head
		then: [
			(self database at: arg head) == bList 
				ifTrue: [self database at: arg head put: bList tail].
			aList tail: bList tail.
			assoc value value].
	aList := bList.
	aList ifNil: [^self]]! !

!Prolog methodsFor: 'database' stamp: 'spfa 10/3/2022 15:52'!
primlogicRetractall: assoc
        "Purge a database identified by the
         structure in the argument."
	| arg |
	self check: assoc forArity: 1.
	arg := self first: assoc.
	arg isSymbol ifFalse: [self error: 'argument is not a Symbol'].
	self database at: arg put: nil.
	^assoc value! !

!Prolog methodsFor: 'as yet unclassified' stamp: 'spfa 10/19/2022 23:28'!
primlogicRule: assoc

"not yet implemented"

	self error: 'Rules can only be compiled as methods'! !

!Prolog methodsFor: 'Smalltalk interop' stamp: 'spfa 11/18/2009 18:15'!
primlogicRuntime: assoc
        "Unify the first argument with
         the system time in milliseconds."
    self check: assoc forArity: 1.
    self unify: assoc key head
        with: Time millisecondClockValue
        then: [^assoc value]! !

!Prolog methodsFor: 'Smalltalk interop' stamp: 'spfa 9/28/2022 21:25'!
primlogicSend: assoc

	assoc key length < 2 ifTrue:
		[self check: assoc forArity: 2].

    assoc key length = 2 ifTrue: [
    		(self first: assoc) perform: (self second: assoc).
        	^assoc value].

    assoc key length = 3 ifTrue: [
    	self unify: assoc key tail tail head 
			with: ((self first: assoc) perform: (self second: assoc)) 
        	then: [^assoc value].
	^self].

    self check: assoc forArity: 4.

    self unify: assoc key tail tail tail head 
		with: ((self first: assoc) perform: (self second: assoc) 
			withArguments: (self third: assoc) topForms)  
        	then: [^assoc value]

! !

!Prolog methodsFor: 'higher-order predicates' stamp: 'spfa 10/13/2022 16:24'!
primlogicSetof: assoc

	"No support for existential quantifiers at the moment"

    	| goal answers template bag mark cachedVarList |

   	self check: assoc forArity: 3.

	template := assoc key head.
	template isPrologList ifFalse: [template := nil].
	bag := OrderedCollection new.

	goal := self getGoal: (self second: assoc).

	cachedVarList := varList copy.
	self addVarFrom: assoc key. 
	mark := unboundVars size.

	answers := self doGoal: goal 
					onSuccess: [
						self inferAnswer: false.
						answer first ifNotEmpty: [
							bag add: (template 
									ifNotNil: [template resultCopy] 
									ifNil: [answer last first logicResult])]].
	[answers isClosure] whileTrue: [answers := answers value].

	bag ifEmpty: [^ self "fails"].

	assoc key tail tail head getValueFrom: bag asSet asArray asPrologList.

	varList := cachedVarList.
	self unbindVar: mark. 
	answer := OrderedCollection new.

	^ assoc value! !

!Prolog methodsFor: 'Smalltalk interop' stamp: 'spfa 10/4/2022 22:15'!
primlogicTest: assoc
        "Succeed if the argument is true, fail if it is false, raise an error if not a Boolean."
	| bool |

   	self check: assoc forArity: 1.
	((bool := self first: assoc) isKindOf: Boolean) ifFalse: [self error: 'not a Boolean'].
    	bool ifTrue: [^assoc value]! !

!Prolog methodsFor: 'success-failure' stamp: 'spfa 11/18/2009 18:15'!
primlogicTrue: assoc
        "Succeed always."
    assoc key ifNotNil: [^self].
    ^assoc value! !

!Prolog methodsFor: 'basics' stamp: 'spfa 9/29/2022 22:21'!
primlogicUnify: assoc
        "Succeed if the two arguments can be unified."
    self check: assoc forArity: 2.
    self unify: assoc key head
        with: assoc key tail head
        then: [^assoc value]! !

!Prolog methodsFor: 'classifying terms' stamp: 'spfa 9/26/2022 14:06'!
primlogicVar: assoc
        "Succeed if the argument is an
         unbound variable."
    self check: assoc forArity: 1.
    (self first: assoc) class == LogicUnbound
        ifTrue: [^assoc value].
    (self first: assoc) class == LogicDontCare
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'classifying terms' stamp: 'spfa 9/28/2022 22:13'!
primlogicWrite: assoc
        "Write all the arguments to the output stream."
    | value |
    assoc key do: [:argument |
        value := argument logicInstanciated.
        self outputStream nextPutAll:
            ((value isKindOf: String)
                ifTrue: [value]
                ifFalse: [value prologPrintString]).
	   self finalizeOutput].
    ^assoc value! !

!BlockClosure methodsFor: '*Prolog-Core' stamp: 'spfa 10/13/2022 00:56'!
printForPrologOn: aStream

	| ds |
	
	ds := self decompile decompileString.

	aStream nextPutAll: (ds copyFrom: 2 to: ds size - 1)! !

!Prolog methodsFor: 'interaction' stamp: 'spfa 11/18/2009 18:15'!
prompt

	^ self class asString, ' ?-'! !

!Prolog class methodsFor: 'database' stamp: 'spfa 9/26/2022 09:10'!
purge: aSymbol

	self database purge: aSymbol! !

!Prolog methodsFor: 'database' stamp: 'spfa 9/26/2022 09:10'!
purge: aSymbol

	self database purge: aSymbol! !

!Prolog methodsFor: 'evaluation' stamp: 'spfa 10/13/2022 16:46'!
query: anObject 

	^ self generatorOnQuery: anObject withBindings: true! !

!Prolog class methodsFor: 'predicates' stamp: 'spfa 10/19/2022 23:15'!
readPredicate: aString

	"Predicates are encoded in different forms - for example true can appear as 'true', 'logicTrue', or 'primlogicTrue' depending on the parsing context"

	| symb |

	symb := aString asSymbol.

^ self binaryPredicates at: symb ifAbsent: [

	 self predicates at: symb ifAbsent: [ | logicSymb binaryLogicSymb primlogicSymb pred |

		(symb beginsWith: ('prim', self prologKeyword)) ifTrue: 
			[self predicates at: symb put: symb. ^ symb].

		(symb beginsWith: self prologKeyword) ifTrue: [
			symb := (symb allButFirst: self prologKeyword size) 
							withFirstCharacterDownshifted].

		symb last == $: ifTrue: [symb := symb allButLast].
		symb := symb asSymbol.

		logicSymb := self makeLogicSelector: symb.	
		primlogicSymb := ('prim', logicSymb, ':') asSymbol.

		pred := (self canUnderstand: logicSymb)
			ifTrue: [logicSymb] ifFalse: [
				(self canUnderstand: primlogicSymb)
					ifTrue: [primlogicSymb] ifFalse: [symb]].

		self predicates at: symb put: pred.
		self predicates at: logicSymb put: pred.
		pred == primlogicSymb
			ifTrue: [self predicates at: primlogicSymb put: pred].

		^ pred]]
! !

!PrologInfo methodsFor: 'as yet unclassified' stamp: 'spfa 10/13/2022 18:42'!
readmeText
	"Text shown to the user after load"
	| s |
	s := String new writeStream.
	s nextPutAll: '
This is Prolog for Squeak, initially adapted by Bolot Kerimbaev from Prolog/V by Mike Teng,
now being refactored by Stephane Rollandin for Squeak 5.1 and newer.

For old-style Prolog workspace:
	PrologWorkspace openDemo.   "doIt"

For newer integrated browser & REPL:
	Family new openGui.		"doIt"
	Chapter2 new openGui.		"doIt"

The original paper from Mike Teng should be available at
	http://www.zogotounga.net/comp/squeak/v-prolog.pdf
...note that this is largely deprecated.

Home page:
	http://www.zogotounga.net/comp/squeak/prolog.htm



produced from package ';
		nextPutAll: self packageName;
		nextPutAll: ' on ';
		nextPutAll: Date today printString;
		space;
		nextPutAll: Time now printString;
		cr.
	^s contents.! !

!Prolog class methodsFor: 'prolog-compilation' stamp: 'spfa 11/18/2009 18:15'!
removeSelector: selector 
	
	super removeSelector: selector .
	self withAllSubclassesDo: [:pl | pl clearPredicates]! !

!PrologBrowser methodsFor: 'as yet unclassified' stamp: 'spfa 8/16/2009 23:04'!
repl

	^ repl ifNil: [repl _ PrologREPLMorph on: Prolog new]! !

!PrologBrowser methodsFor: 'as yet unclassified' stamp: 'spfa 11/18/2009 18:15'!
repl: aREPLMorph

	repl := aREPLMorph! !

!Prolog class methodsFor: 'database' stamp: 'spfa 11/18/2009 18:15'!
retract: anArray

	self new primlogicRetract: (PrologList with: anArray asPrologList) -> []
	! !

!Prolog methodsFor: 'database' stamp: 'spfa 9/26/2022 09:17'!
retract: anArray

	self primlogicRetract: (PrologList with: anArray asPrologList) -> []
	! !

!Prolog class methodsFor: 'predicates' stamp: 'spfa 9/26/2022 17:02'!
selectorForBinaryPredicate: aSymbol

	^ self binaryPredicates at: aSymbol ifAbsent: [self error: 'Undefined binary predicate: ', aSymbol]! !

!PrologREPLMorph methodsFor: 'as yet unclassified' stamp: 'spfa 10/13/2022 17:15'!
setMaxAnswers

	self maxAnswers: (FillInTheBlank request: 'Maximum number of displayed answers') asNumber! !

!PrologWorkspace class methodsFor: 'as yet unclassified' stamp: 'spfa 12/1/2011 10:28'!
shouldStyle

	^ false! !

!Prolog class methodsFor: 'prolog-compilation' stamp: 'spfa 11/30/2011 14:48'!
shoutParserClass

	self == Prolog ifTrue: [^ super shoutParserClass].

	^ SHDummyParser! !

!Prolog class methodsFor: 'prolog-compilation' stamp: 'spfa 11/18/2009 18:15'!
sourceCodeTemplate
	"Answer an expression to be edited and evaluated in order to define 
	methods in this class."

    ^ Logic prologKeyword, 'Predicate
"comment"
predicate(term1, term2) :-
    goal1(term3),
    goal2(term4).'! !

!PrologBrowser methodsFor: 'as yet unclassified' stamp: 'spfa 12/1/2011 13:19'!
wantsAnnotationPane

	^ false! !

!Prolog class methodsFor: 'prolog-compilation' stamp: 'bolot 7/12/1999 19:22'!
wantsChangeSetLogging
	^true! !

!PrologBrowser methodsFor: 'as yet unclassified' stamp: 'spfa 9/28/2022 22:17'!
wantsOptionalButtons

	^ true! !
