SystemOrganization addCategory: #'Prolog-Tests'!


Prolog subclass: #PrologCLDForTesting
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Tests'!

Prolog subclass: #PrologILDForTesting
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Tests'!

PrologCLDForTesting class
	instanceVariableNames: ''!

TestCase subclass: #PrologTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Tests'!

PrologTest class
	instanceVariableNames: ''!

PrologILDForTesting class
	instanceVariableNames: ''!


!PrologCLDForTesting commentStamp: 'spfa 6/20/2008 10:49' prior: 0!
A Prolog with class-local database

assert & retract effect all instances of this class, but leave the global database untouched!
]style[(128)!

!PrologILDForTesting commentStamp: 'spfa 6/20/2008 10:55' prior: 0!
A Prolog with instance-local database

assert & retract expression effect only their interpreter instance 
programs maintaining a temporary database are assured not to pollute any global state (see for example the animals game in Chapter2)

note that the stability of programs interactively intered in the REPL that relies on the database is very brittle since any change of class in the browser will delete the current interpreter instance and thus reset the database!
]style[(468)!

!PrologCLDForTesting class methodsFor: 'as yet unclassified' stamp: 'spfa 11/18/2009 18:15'!
databaseScope

	^ #classLocal! !

!PrologILDForTesting class methodsFor: 'as yet unclassified' stamp: 'spfa 11/18/2009 18:15'!
databaseScope

	^ #instanceLocal! !

!PrologTest methodsFor: 'failures' stamp: 'spfa 10/20/2022 13:02'!
expectedFailures

	^ #(testLength testBagof)! !

!PrologILDForTesting methodsFor: 'as yet unclassified' stamp: 'spfa 10/3/2022 21:42'!
logicBelieves

believes(john, likes(mary, pizza)).
believes(frank, likes(mary, fish)).
believes(john, likes(mary, apples)).! !

!PrologCLDForTesting methodsFor: 'as yet unclassified' stamp: 'spfa 10/4/2022 00:04'!
logicDoUniv
"
a(1).
b(1,2).
"
doUniv(FUNCTOR, ARGS) :-
  univ(GOAL,[FUNCTOR|ARGS]),
  call(GOAL).

! !

!PrologCLDForTesting methodsFor: 'as yet unclassified' stamp: 'spfa 10/4/2022 00:08'!
logicDountilstop

dountilstop :-
  repeat,
  readstring(X),
  ==(X, 'stop'), !!.! !

!PrologILDForTesting methodsFor: 'as yet unclassified' stamp: 'spfa 10/3/2022 21:42'!
logicItalian

italian(marco).! !

!PrologILDForTesting methodsFor: 'as yet unclassified' stamp: 'spfa 10/3/2022 21:42'!
logicLikes

likes(mary, pizza).
likes(marco, pizza).
likes(PERSON, pizza) :- italian(PERSON).! !

!PrologILDForTesting methodsFor: 'as yet unclassified' stamp: 'spfa 10/3/2022 22:25'!
logicSayHello

"see related test case"

sayHello([_,X]) :- 
	=(X,'HELLO').! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/4/2022 01:15'!
testAppend

"from 'Fifty Years of Prolog and Beyond' (2022) - https://arxiv.org/abs/2201.10816

Technical Test: 
	As a rule of thumb, one can check if a logic solver can be considered as a Prolog system or not via the following test. The test requires that the well-known append/3 predicate can be written exactly as follows: 

	append([], X, X). 
	append([H | X], Y, [H | Z]) :- append(X, Y, Z). 

and can be queried in variety of ways, e.g., 

	to append two lists: 
		append([1, 2], [c, d], R)

	deconstruct a list as in: 
		append(A, B, [1, 2])

	or with any arbitrary instantiation of the arguments, such as 
		append([X | T], [c], [Z, Z, Z])
"

	self assert: (Prolog ? 'append([1, 2], [c, d], R)') printString = '{{[1, 2, c, d]}}'.

	self assert: (Prolog ? 'append(A, B, [1, 2])') printString = '{{nil . [1, 2]} . {[1] . [2]} . {[1, 2] . nil}}'.

	self assert: (Prolog ? 'append([X | T], [c], [Z, Z, Z])') printString = '{{#c . [c] . #c}}'.

	"see Prolog>>#logicAppend for the implementation"
! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 23:22'!
testArg

	| v |

	{
		'arg(2,f(3,4,5,6),X)' -> #((4)).
		'arg(N,f(3,8,9,20,4,5,6),4)' -> #((5)).
		'arg(2,f(3,N,5,6),7)' -> #((7)).
	}
	do: [:assoc | 
		self assert: (Prolog evaluate: assoc key) = assoc value].

	v := Prolog evaluate: 'arg(2,f(3,N,5,6),X)' withBindings: true.
	self assert: v printString = '{{N = X.}}'! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 22:05'!
testArg2

	{
		'=(X,f(1,2,3)), arg(2,X,V)' -> '(f(1, 2, 3) 2)'.
		'=(X,f(3,2,6)), arg(N,X,3)' -> '(f(3, 2, 6) 1)'.
		'=(X,f(3,U,6)), arg(2,X,27)' -> '(f(3, 27, 6) 27)'.
	}
	do: [:assoc | 
		self assert: (String streamContents: [:str | 
					(Prolog evaluate: assoc key) first printElementsOn: str]) = assoc value].
! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 22:07'!
testArg3

	self assert: (Prolog evaluate: 'arg(2, test(X,Z,Y), hello)') first first = #hello.

	self assert: (Prolog evaluate: 'arg(1, connect(yard, pen), X)') = #(#(#yard)).! !

!PrologTest methodsFor: 'arithmetic' stamp: 'spfa 10/14/2022 11:04'!
testArithmetic

	self assert: (Prolog new evaluate: '5 == (4 + 1)') isNil.

	self assert: (Prolog new evaluate: 'N is (4 + 1), 5 == N') notNil.! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/11/2022 22:00'!
testAtom

	self assert: (Prolog ? 'atom(blep)') = #(()).

	self assert: (Prolog ? 'atom(5)') isNil.
	self assert: (Prolog ? 'atom(''blep'')') isNil.
	self assert: (Prolog ? 'atom([])') isNil.
	self assert: (Prolog ? 'atom(blep())') isNil.
! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/11/2022 22:01'!
testAtomic

	self assert: (Prolog ? 'atomic(blep)') = #(()).
	self assert: (Prolog ? 'atomic([])') = #(()).
	self assert: (Prolog ? 'atomic(''blep'')') = #(()).
	self assert: (Prolog ? 'atomic(5)') = #(()).

	self assert: (Prolog ? 'atomic(blep())') isNil.
! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/11/2022 22:02'!
testAtomic2

	self assert: (Prolog ? 'iss(M,Morph new),atomic(M)') notNil! !

!PrologTest methodsFor: 'printing' stamp: 'spfa 10/13/2022 11:41'!
testAutoReferentialPrinting

	| x y |

	x := LogicRef newTerm: 'X'.
	x getValueFrom: (PrologList with: x).

	self assert: x isAutoReferential.
	self assert: (x isAutoReferentialWith: (Set with: x)).

	self assert: x prologPrintString = 'X = [X].'.

	y := LogicRef newTerm: 'Y'.
	y getValueFrom: ((PrologList with: x) tail: y).

	self assert: y prologPrintString = 'Y = [X|Y].'.
! !

!PrologTest methodsFor: 'printing' stamp: 'spfa 10/13/2022 11:01'!
testAutoReferentialPrinting2

	| x y |

	x := LogicRef newTerm: 'X'.
	x getValueFrom: {#zo . x} asPrologRelation.

	self assert: x prologPrintString = 'X = zo(X).'.

	y := LogicRef newTerm: 'Y'.
	y getValueFrom: {#bu . x . y} asPrologRelation.

	self assert: y prologPrintString = 'Y = bu(X, Y).'.! !

!PrologTest methodsFor: 'higher-order' stamp: 'spfa 10/3/2022 21:44'!
testBagof

"----------- no support for existential quantifier - bagof/3 semantics is not respected"
	| results |

	results := PrologILDForTesting evaluate: 'bagof(likes(mary, X), believes(_, likes(mary, X)), BAG)' 
					getBindings: 'BAG'.

	self assert: results size = 3.

	self assert: results first printString = '[likes(mary, pizza)]'.
	self assert: results second printString = '[likes(mary, fish)]'.
	self assert: results last printString = '[likes(mary, apples)]'.

"------------"

	self assert: 
		(PrologILDForTesting evaluate: 'bagof(P, likes(P, pizza), B)' 
					getBindings: 'B') 
		last printString = '[mary, marco, marco]'.! !

!PrologTest methodsFor: 'higher-order' stamp: 'spfa 9/28/2022 00:29'!
testBagof2

	self assert: 
		(PrologILDForTesting evaluate: 'bagof(person, likes(person, ''wine''), bag)') isNil! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 21:56'!
testBasics

	self assert: (Prolog new evaluate: '=([3,4,5],[X|Y])') = {{3 . {4 . 5} asPrologList}}.
	self assert: (Prolog new evaluate: '=([X,4,5],[3|Y])') = {{3 . {4 . 5} asPrologList}}.

	self assert: (Prolog new evaluate: '=([3,4,5],[X,Y|Z])') = {{3 . 4 . {5} asPrologList}}.
	self assert: (Prolog new evaluate: '=([3,4,5],[X,Y,Z|T])') = #((3 4 5 nil)).

	self assert: (Prolog new evaluate: '=([X|Y],[3,4,5])') = {{3 . {4 . 5} asPrologList}}.
	self assert: (Prolog new evaluate: '=([3|Y],[X,4,5])') = {{{4 . 5} asPrologList . 3}}.

	self assert: (Prolog new evaluate: '=([X,Y|Z],[3,4,5])') = {{3 . 4 . {5} asPrologList}}.
	self assert: (Prolog new evaluate: '=([X,Y,Z|T],[3,4,5])') = #((3 4 5 nil)).! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 21:57'!
testBasics1

	self assert: (Prolog new evaluate: '=([3,Y,5],[X,4|Z])') = {{4 . 3 . {5} asPrologList}}.
	self assert: (Prolog new evaluate: '=([3,4,Z],[X,Y,5|T])') = #((5 3 4 nil)).

	self assert: (Prolog new evaluate: '=([X,4|Z],[3,Y,5])') = {{3 . {5} asPrologList . 4}}.
	self assert: (Prolog new evaluate: '=([X,4,Z|T],[3,Y,5])') = #((3 5 nil 4)).! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 22:13'!
testBasics2

	self assert: (Prolog evaluate: '=([U,b],[a,V])') = #((a b))! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/14/2022 10:56'!
testBetween

	| gen |

	self assert: (Prolog new evaluate: 'between(5,8,7)') = #(()).
	self assert: (Prolog new evaluate: 'between(5,8,10)') isNil.

	gen := Prolog new query: 'between(5,9,N)'.
	self assert: (gen upToEnd asArray collect: #printString)
		= #('{N = 5.}' '{N = 6.}' '{N = 7.}' '{N = 8.}' '{N = 9.}').

	gen := Prolog new query: 'between(5,infinity,N)'.
	self assert: ((gen next: 7) asArray collect: #printString)
		= #('{N = 5.}' '{N = 6.}' '{N = 7.}' '{N = 8.}' '{N = 9.}' '{N = 10.}' '{N = 11.}').
	self deny: gen atEnd! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 16:27'!
testBooleans

	self assert: (Prolog ? #(#'true')) = #(()).
	self assert: (Prolog ? #(fail)) isNil.

	self assert: (Prolog ? 'or(true,true)') = #(()).
	self assert: (Prolog ? 'or(true(),fail())') = #(()).
	self assert: (Prolog ? 'or(true,fail)') = #(()).
	self assert: (Prolog ? 'or(fail(),fail())') isNil.
	self assert: (Prolog ? 'or(fail(),true())') = #(()).

	self assert: (Prolog ? 'and(true(),true())') = #(()).
	self assert: (Prolog ? 'and(true(),fail())') isNil.
	self assert: (Prolog ? 'and(fail(),fail())') isNil.
	self assert: (Prolog ? 'and(fail(),true())') isNil.

	self assert: (Prolog ? 'true,true') = #(()).
	self assert: (Prolog ? 'true,fail') isNil.
	self assert: (Prolog ? 'fail,fail') isNil.
	self assert: (Prolog ? 'fail,true') isNil.
! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 22:40'!
testBooleans2

	"Beware that in Prolog, true is a predicate, not the Smalltalk boolean"

	self assert: (Prolog ? 'equal(true,true)') = #(()).
	self assert: (Prolog ? 'equal(5,5)') = #(()).

	self assert: (Prolog ? 'equal(eq(N,5),eq(M,5))') isNil.
"	self assert: (Prolog ? 'equal(eq(5,5),eq(5,5))') isNil." "implementation detail?"
	self assert: (Prolog ? 'equal(eq(5,5),true)') isNil.

	self should: [Prolog ? 'test(true)'] raise: Error.

	! !

!PrologTest methodsFor: 'interaction' stamp: 'spfa 11/18/2009 18:15'!
testBrowser

	self assert: Prolog new gui isMorph.
! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/13/2022 09:41'!
testCompilation

	Metalevel compileAll! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/13/2022 09:42'!
testCompilation2

	Chapter5 compileAll! !

!PrologTest methodsFor: 'database' stamp: 'spfa 10/3/2022 21:48'!
testDatabase

	{
		'retractall(brother).' 
			-> #(()).
		'assertz(brother(john, stewart)),
					asserta(brother(john, barbara)), assertz(brother(john, alan)).'
			-> #(()).
		'brother(X,Y)'
			->  #((john barbara) (john stewart) (john alan)).
		'retract(brother(john, stewart))'
			-> #(()).
		'brother(X,Y)'
			->  #((john barbara) (john alan)).
		'retractall(brother).' 
			-> #(()).
	}
	do: [:assoc | 
		self assert: (Family evaluate: assoc key) = assoc value]! !

!PrologTest methodsFor: 'higher-order' stamp: 'spfa 10/4/2022 22:08'!
testDoUniv

	self assert: 
		(PrologCLDForTesting evaluate: 'retractall(a),assert(a(1)), doUniv(a,[X])') 
			printString = '#(#(1))'.

	self assert: 
		(PrologCLDForTesting evaluate: 'retractall(b),assert(b(1,2)), doUniv(b,[X,Y])') 
			printString = '#(#(1 2))'.! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 22:14'!
testFamily

	{
		'father(X,Y)' 
			-> #(('John' 'Mary') ('John' 'David') ('David' 'Jack') ('Arthur' 'Nancy')).
		'grandPa(''John'',Y)' 
			-> #(('Nancy') ('Jack')).
		'father(X, ''Mary'')'
			-> #(('John')).
		'grandPa(''John'', X), father(Y, X)'
			->  #(('Nancy' 'Arthur') ('Jack' 'David')).
		'and(grandPa(''John'', X), father(Y, X))'
			->  #(('Nancy' 'Arthur') ('Jack' 'David')).
		'and(grandPa(''John'', X), father(Y, X),mother(Z,X))'
			->  #(('Nancy' 'Arthur' 'Mary')).
		'grandPa(''John'', X), father(Y,X),mother(Z,X)'
			->  #(('Nancy' 'Arthur' 'Mary')).
		'father(''John'', ''Mary'')'
			-> #(()).
		'health(''John'', Y)'
			->  #(('John had two times of heart attacks')).
		'consult( healthHistory(''John'', X), Doctor new)'
			->  #(('John had two times of heart attacks')).
	}
	do: [:assoc | self assert: (Family evaluate: assoc key) = assoc value]! !

!PrologTest methodsFor: 'problematic' stamp: 'spfa 9/29/2022 01:03'!
testFamily2

"
	At the moment, a database entry with a term also defined as a hard-coded predicate will be completely ignored.
	See Logic>>#doesNotUnderstand:
"

	Family ? 'assert(father(''Paul'',''Nicole''))'.
	
	self deny: (Family ? 'father(''Paul'',c)') = #(('Nicole'))! !

!PrologTest methodsFor: 'higher-order' stamp: 'spfa 10/3/2022 22:08'!
testFindall

	self assert: (Chapter2 evaluate: 'perm([1,2,3],X)' getBindings: 'X') 
		last printString = '[3, 2, 1]'.

	self assert: (Chapter2 evaluate: 'findall(X,perm([1,2,3],X),All)' getBindings: 'All') 
		last printString = '[[1, 2, 3], [2, 1, 3], [2, 3, 1], [1, 3, 2], [3, 1, 2], [3, 2, 1]]'
! !

!PrologTest methodsFor: 'higher-order' stamp: 'spfa 10/3/2022 22:10'!
testFindall2

	self assert: 
		(PrologILDForTesting evaluate: 'findall(likes(mary, X), believes(_, likes(mary, X)), Bag)' 
					getBindings: 'Bag') 
			last printString = '[likes(mary, pizza), likes(mary, fish), likes(mary, apples)]'.


	self assert: 
		(PrologILDForTesting evaluate: 'findall(Person, likes(Person, pizza), Bag)' getBindings: 'Bag') 
			last printString = '[mary, marco, marco]'.
	
	self assert: 
		(PrologILDForTesting evaluate: 'findall(Person, likes(Person, wine), Bag)' getBindings: 'Bag') 
			last prologPrintString = '[]'.! !

!PrologTest methodsFor: 'higher-order' stamp: 'spfa 10/3/2022 22:11'!
testFindall3

	self assert: 
		(PrologILDForTesting evaluate: 'findall([X,Y],believes(X,Y),All)' 
					getBindings: 'All') 
			last printString 
	= '[[john, likes(mary, pizza)], [frank, likes(mary, fish)], [john, likes(mary, apples)]]'! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/11/2022 10:38'!
testFisherCh2

	{
		"http://www.csupomona.edu/~jrfisher/www/prolog:=tutorial/2:=2.html"

		'factorial(3,W)' -> #((6)).
		'factorialT(5,1,F)' -> #((120)).

		"http://www.csupomona.edu/~jrfisher/www/prolog:=tutorial/2:=7.html"

		'member(2,[1,2,3])' -> #(()).
		'member(X,[1,2,3])' -> #((1) (2) (3)).
		'member([3,Y], [[1,a],[2,m],[3,z],[4,v],[3,p]])' -> #((z) (p)).
		'member(X,[23,45,67,12,222,19,9,6]), is(Y, X*X), lt(Y, 100)'
			-> #((9 81) (6 36)).
	}
	do: [:assoc | 
		self assert: (Chapter2 evaluate: assoc key) = assoc value].


	{
		'takeout(X,[1,2,3],L)' ->  '{{1 . [2, 3]} . {2 . [1, 3]} . {3 . [1, 2]}}'.
		'takeout(3,W,[a,b,c])' ->  '{{[3, a, b, c]} . {[a, 3, b, c]} . {[a, b, 3, c]} . {[a, b, c, 3]}}'.
	}
	do: [:assoc | 
		self assert: (Chapter2 evaluate: assoc key) printString = assoc value].


! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 22:16'!
testFisherCh214

	self assert: (String streamContents: [:str |
					(Chapter2 new outputStream: str) 
						evaluate: 'parse([b,b,a,a,b,a,b])'])
=
'0  [b, b, a, a, b, a, b]
0  [b, a, a, b, a, b]
0  [a, a, b, a, b]
1  [a, b, a, b]
1  [b, a, b]
2  [a, b]
2  [b]
2  []
'.

	self assert: (String streamContents: [:str |
					(Chapter2 new outputStream: str) 
						evaluate: 'parse([b,b,a])'])
		=
'0  [b, b, a]
0  [b, a]
0  [a]
'! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 22:11'!
testFisherCh27

	self assert: ((Chapter2 evaluate: 'perm([1,2,3],X)') 
					collect: [:ea | ea first arrayForm])
		= #(#(1 2 3) #(2 1 3) #(2 3 1) #(1 3 2) #(3 1 2) #(3 2 1))! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 22:16'!
testFisherCh28

	self assert: (Chapter2 evaluate: 'change([0,3,D,N,P])') 
		= #(#(0 0 25) #(0 1 20) #(0 2 15) #(0 3 10) #(0 4 5) #(0 5 0) #(1 0 15) #(1 1 10) #(1 2 5) #(1 3 0) #(2 0 5) #(2 1 0)).

	self assert: (Chapter2 evaluate: 'change([0,2,3,2,P]).') = #(#(10)).! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 22:17'!
testFisherCh28Extra

"an extra test, for the =/2 predicate"

	self assert: (Chapter2 evaluate: '=(N,2),change([0,N,Q,N,P]).')
		= #(#(2 0 40) #(2 1 30) #(2 2 20) #(2 3 10) #(2 4 0)).

"and one more for is/2"

	self assert: (Chapter2 evaluate: 'is(N,1+1),change([0,N,Q,N,P]).')
		= #(#(2 0 40) #(2 1 30) #(2 2 20) #(2 3 10) #(2 4 0)).! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/11/2022 10:44'!
testFisherCh3

	{
		"http://www.csupomona.edu/~jrfisher/www/prolog:=tutorial/3:=1.html"

		'=(p(X,f(Y),a), p(a,f(a),Y))' -> #((a a)).
		'=(p(X,f(Y),a),p(a,f(b),Y))' -> nil.
		'=(p(X,f(Y),a) , p(Z,f(b),a))' ->  #((b)).
		'=(p(X,f(Y),a) , p(Z,f(b),a)), =(X, d) ' -> #((d b d)).

		"http://www.csupomona.edu/~jrfisher/www/prolog:=tutorial/3:=2.html"

		'p(X)' -> #((a) (a) (b) (d)).
		'p(X),!!' -> #((a)).
		'r(X),!!,s(Y)' -> #((a a) (a b) (a c)).
		'r(X), s(Y), !!' -> #((a a)).

		'color(a,C)' -> #(('red')).
		'color(c,C)' -> #(('unknown'))
	}
	do: [:assoc | 
		self assert: (Chapter3 evaluate: assoc key) = assoc value]! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/11/2022 22:49'!
testFunctor

	self assert: (Prolog evaluate: 'functor(father(3,4),H,A)') 
					printString = '#(#(#father 2))'.

	self assert: (Prolog evaluate: 'functor(father(3,4),father,A)') = #((2)).
	self assert: (Prolog evaluate: 'functor(father(3,4),father,2)') = #(()).
	self assert: (Prolog evaluate: 'functor(father(3,4),S,2)') = #((father)).

	self assert: (Prolog evaluate: 'functor(F,houba,3)') 
					= {{{#houba . #'_' . #'_' . #'_'} asPrologRelation}}
	! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 22:04'!
testFunctor2

	self assert: (Prolog evaluate: 'functor(T, connect, 2), arg(1, T, yard), arg(2, T, pen).')  
					printString = '{{connect(yard, pen)}}'.
! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 21:53'!
testGround

	self assert: (Prolog ? 'ground(a)') = #(()).

	self assert: (Prolog ? 'ground(=(H,a))') isNil.
	self assert: (Prolog ? 'ground(=(X))') isNil.

	self assert: (Prolog ? 'ground([1,2,3,4])') = #(()).
	self assert: (Prolog ? 'ground([1,2,Z,4])') isNil.
	self assert: (Prolog ? 'ground([1,2,3,Z])') isNil.
	self assert: (Prolog ? 'ground([1,2,[3,Z]])') isNil.
	self assert: (Prolog ? 'ground([1,2,3 | Z])') isNil.
	self assert: (Prolog ? 'ground([1,2,[3|Z]])') isNil.

	self assert: (Prolog ? '=(A,42),ground(A)') = #(#(42)).
	self assert: (Prolog ? '=(A,X),ground(A)') isNil.
	self assert: (Prolog ? '=(A,X),ground(ground(A))') isNil.! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/8/2022 00:30'!
testGround2

	self assert: (Prolog ? 'ground(f(x,y))') = #(()).
	self assert: (Prolog ? 'ground(f(x,_))') isNil.

	self assert: (Prolog ? '=(K,f(X,Y)), \+(ground(f(X,Y))), =(X,x), =(Y,y), ground(f(X,Y)).') 
			printString = '{{f(x, y) . #x . #y}}'.! !

!PrologTest methodsFor: 'parsing' stamp: 'spfa 10/20/2022 12:12'!
testHomoiconicity

	self assert: (Prolog ? ' L = (X=Y,Z=Q),functor(L,F,A).') printString = '{{,(=(X, Y), =(Z, Q)) . #, . 2}}'! !

!PrologTest methodsFor: 'parsing' stamp: 'spfa 10/20/2022 12:13'!
testHomoiconicity2

	self assert: (Prolog ? 'L = (ten(X):-(X is 2*5))') printString = '{{:-(ten(X), is(X, *(2, 5)))}}'! !

!PrologTest methodsFor: 'parsing' stamp: 'spfa 10/20/2022 12:14'!
testHomoiconicity3

	self assert: (Prolog ? 'L = (sentence --> noun, verbPhrase)') printString = '{{-->(sentence, ,(noun, verbPhrase))}}'! !

!PrologTest methodsFor: 'interaction' stamp: 'spfa 10/13/2022 16:42'!
testInferenceSuspension

	| gen |

	gen := Chapter2 new generatorOnQuery: 'perm([1,2,3],X)' withBindings: true.

	gen next: 3.

	self assert: gen next printString = '{X = [1, 3, 2].}'.

	self assert: gen upToEnd last printString =  '{X = [3, 2, 1].}'.

! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/12/2022 14:25'!
testInfix

	self assert: ((Prolog new evaluate: 'X = 1' withBindings: true))
					printString = '{{X = 1.}}'.

	self assert: ((Prolog new evaluate: '1 = X' withBindings: true))
					printString = '{{X = 1.}}'.

	self assert: ((Prolog new evaluate: 'X = 1, Y = 2' withBindings: true)) 
					printString = '{{X = 1. . Y = 2.}}'.
! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/12/2022 14:25'!
testInfix1b

	self assert: ((Prolog new evaluate: '1 = X, X = Y' withBindings: true))
					printString = '{{X = 1. . Y = 1.}}'.! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/10/2022 14:10'!
testInfix2

	self assert: ((Prolog new evaluate: '[X,Y] = [1,Z]' withBindings: true))		 
					printString = '{{X = 1. . Y = Z.}}'.! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/10/2022 16:39'!
testInfix3

	self assert: ((Prolog new evaluate: 'univ(X = 1,L)' withBindings: true))
					printString = '{{L = [=, X, 1].}}'..
! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/10/2022 16:39'!
testInfix4

	self assert: ((Prolog new evaluate: 'univ(bah = X,L)' withBindings: true))
					printString = '{{L = [=, bah, X].}}'.! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/11/2022 00:16'!
testInfix5

	self assert: ((Prolog new evaluate: 'X = (Y = 1)' withBindings: true)) 
					printString = '{{X = =(Y, 1).}}'.

	self assert: ((Prolog new evaluate: '(Y = 1) = X' withBindings: true)) 
					printString = '{{X = =(Y, 1).}}'.

	self assert: ((Prolog new evaluate: '(X = Y) = Z' withBindings: true)) 
					printString = '{{Z = =(X, Y).}}'.! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/11/2022 15:58'!
testInfix5b

	self assert: ((Prolog new evaluate: 'X = =(Y,1).' withBindings: true)) 
					printString = '{{X = =(Y, 1).}}'.

	self assert: ((Prolog new evaluate: '=(Y,1) = X' withBindings: true)) 
					printString = '{{X = =(Y, 1).}}'.

	self assert: ((Prolog new evaluate: '=(X,Y) = Z' withBindings: true)) 
					printString = '{{Z = =(X, Y).}}'.! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/11/2022 15:55'!
testInfix6

	self assert: ((Prolog new evaluate: 'X is 5 + 2' withBindings: true))  
					printString = '{{X = 7.}}'.! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/10/2022 23:15'!
testInfixPrecedence

	self should: [Prolog new evaluate: 'X = 1 =** List'] raise: Error.

	self assert: ((Prolog new evaluate: '(X=1) =** List' withBindings: true))
					printString = '{{List = [=, X, 1].}}'.! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/10/2022 23:15'!
testInfixPrecedence2
 
	self assert: ((Prolog new evaluate: 'X = (1 =** List)' withBindings: true))
					printString = '{{X = =**(1, List).}}'.! !

!PrologTest methodsFor: 'database' stamp: 'spfa 9/26/2022 09:20'!
testInstanceDatabases

	| prolog |

	prolog := PrologILDForTesting new.

	prolog assert: #(brother 'Yann' 'Paul').
	self assert: (prolog ? #(brother ~ ~))  = #(('Yann' 'Paul')).

	self should: [PrologILDForTesting ? #(brother ~ ~)] raise: Error.

	prolog assert: #(brother 'Henry' 'Jean').
	self assert: (prolog ? #(brother ~ ~))  = #(('Yann' 'Paul') ('Henry' 'Jean')).

	prolog retract: #(brother 'Henry' 'Jean').
	self assert: (prolog ? #(brother ~ ~))  = #(('Yann' 'Paul')).

	prolog assert: #(brother 'Henry' 'Jean').

	prolog retract: #(brother ~ ~).

	self assert: (prolog ? #(brother ~ ~))  isNil.

	prolog purge: #brother.

	self should: [prolog ? #(brother ~ ~)] raise: Error
! !

!PrologTest methodsFor: 'database' stamp: 'spfa 9/26/2022 09:21'!
testInstanceDatabases2

	| prolog1 prolog2 |

	prolog1 := PrologILDForTesting new.
	prolog2 := PrologILDForTesting new.

	prolog1 assert: #(brother 'Yann' 'Paul').
	self assert: (prolog1 ? #(brother ~ ~))  = #(('Yann' 'Paul')).

	self should: [prolog2 ? #(brother ~ ~)] raise: Error.

	prolog2 assert: #(brother 'Henry' 'Jean').
	self assert: (prolog2 ? #(brother ~ ~))  = #(('Henry' 'Jean')).! !

!PrologTest methodsFor: 'arithmetic' stamp: 'spfa 10/11/2022 10:54'!
testIs1

	self assert: ((Prolog new evaluate: 'X = (5 + 3), Y = (X * 2), is(Z,Y)' withBindings: true)) 
					printString = '{{X = +(5, 3). . Y = *(+(5, 3), 2). . Z = 16.}}'.

	self assert: ((Prolog new evaluate: 'is(N,5)' withBindings: true)) 
					printString = '{{N = 5.}}'.

	self assert: ((Prolog new evaluate: 'is(N, 2+2*5)' withBindings: true)) 
					printString = '{{N = 12.}}'.! !

!PrologTest methodsFor: 'arithmetic' stamp: 'spfa 10/12/2022 21:34'!
testIs2

	self assert: ((Prolog new evaluate: 'X is (((1 + (5 + 4))) / 2)' withBindings: true)) 
					printString = '{{X = 5.}}'.

	self assert: ((Prolog new evaluate: 'univ(X is (((1 + (5 + 4))) / 2), L)' withBindings: true)) 
					printString = '{{L = [is, X, /(+(1, +(5, 4)), 2)].}}'.! !

!PrologTest methodsFor: 'arithmetic' stamp: 'spfa 10/14/2022 11:04'!
testIsVsIss

	self assert: (Prolog new evaluate: 'N is ((4 + 2 * 3) / 2), 5 == N') notNil.

	self assert: (Prolog new evaluate: 'iss(N, ((4 + 2 * 3) / 2)), 5 == N') isNil.
	self assert: (Prolog new evaluate: 'iss(N, ((4 + (2 * 3)) / 2)), 5 == N') notNil.! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/10/2022 11:12'!
testLength

	self assert: (Prolog evaluate: 'length([a, b, c|Hole], L).') notNil! !

!PrologTest methodsFor: 'printing' stamp: 'spfa 10/11/2022 22:49'!
testListPrinting

	self assert: #(1 2 3) asPrologList printString = '[1, 2, 3]'.
	self assert: (#(gloup 2 3) asPrologRelation) printString = 'gloup(2, 3)'.

	self assert: (#(1 2) asPrologList lastCdr: 3) printString = '[1, 2|3]'.

	self assert: (#(1 2) asPrologList lastCdr: LogicDontCare new) printString = '[1, 2|_]'.
	self assert: ({LogicDontCare new. LogicDontCare new} asPrologList 
					lastCdr: LogicDontCare new) printString = '[_, _|_]'.! !

!PrologTest methodsFor: 'database' stamp: 'spfa 9/27/2022 17:16'!
testLocalDatabases

	PrologCLDForTesting purge: #brother.

	PrologCLDForTesting assert: #(brother 'Yann' 'Paul').
	self assert: (PrologCLDForTesting ? #(brother ~ ~))  = #(('Yann' 'Paul')).

	PrologCLDForTesting assert: #(brother 'Henry' 'Jean').
	self assert: (PrologCLDForTesting ? #(brother ~ ~))  = #(('Yann' 'Paul') ('Henry' 'Jean')).

	PrologCLDForTesting retract: #(brother 'Henry' 'Jean').
	self assert: (PrologCLDForTesting ? #(brother ~ ~))  = #(('Yann' 'Paul')).

	PrologCLDForTesting assert: #(brother 'Henry' 'Jean').

	PrologCLDForTesting retract: #(brother ~ ~).

	self assert: (PrologCLDForTesting ? #(brother ~ ~))  isNil.

	PrologCLDForTesting purge: #brother.

	self should: [PrologCLDForTesting ? #(brother ~ ~)] raise: Error
! !

!PrologTest methodsFor: 'core' stamp: 'spfa 9/29/2022 10:24'!
testLogicRef

	self assert: LogicRef new lastNode logicValue class == LogicUnbound.
	self assert: LogicRef new lastNode content class == LogicUnbound.
	self assert: LogicRef new isUnbound.

	! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 9/29/2022 13:59'!
testNegationAnd

	self assert: (Prolog ? 'and(equal(5,6),equal(5,5))') isNil! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 22:54'!
testNegationAsFailure

	self assert: (Prolog ? 'not(equal(5,5))') = nil.

	self assert: (Prolog ? 'not(=(5,5))') = nil.

	self assert: (Prolog ? '\+(=(5,5))') = nil.

	self assert: (Prolog ? 'is(N,5),not(not(==(5,N)))') = {{5}}.
	self assert: (Prolog ? 'is(N,5),not(not(not(==(5,N))))') = nil.


	self assert: (Prolog ? '\+(true)') = nil.
	self assert: (Prolog ? '\+(fail)') = #(()).
! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 9/29/2022 13:59'!
testNegationOr

	self assert: (Prolog ? 'or(equal(5,6),equal(5,5))') = #(()).
! !

!PrologTest methodsFor: 'arithmetic' stamp: 'spfa 10/14/2022 11:29'!
testNegativeNumber

	self assert: (Prolog new evaluate: 'N is 2 + -5') = #(#(-3))! !

!PrologTest methodsFor: 'problematic' stamp: 'spfa 10/11/2022 10:48'!
testNegativeNumberParsing1

	self assert: (Prolog ? '=(X,-100)') = #(#(-100)).
	self assert: (Prolog ? 'is(X,-100)') = #(#(-100)).
	self assert: (Prolog ? 'is(X,10 - 110)') = #(#(-100)).
		
	self assert: (Prolog ? '=(X,100),=(Y,75),is(E,X-Y)') = #(#(100 75 25)).
	self assert: (Prolog ? '=(X,100),=(Y,-75),is(E,X-Y)') = #(#(100 -75 175)).
	self assert: (Prolog ? 'is(X,-100),=(Y,-75),is(E,X-Y)') = #(#(-100 -75 -25)).! !

!PrologTest methodsFor: 'problematic' stamp: 'spfa 10/11/2022 10:49'!
testNegativeNumberParsing2

	self assert: (Prolog ? 'is(X,10-110)') = #(#(-100))! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 21:54'!
testNotNot

	self assert: (Prolog ? '=(X,Y),===(X,Y)') notNil..
	self assert: (Prolog ? 'not(not(=(X,Y))),===(X,Y)') isNil.! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/14/2022 14:03'!
testObfuscated
 
	self assert: (Prolog new evaluate: 'not(\+ (=(1,1),=(2,2),=(1,2)))') isNil.

	self assert: (Prolog new evaluate: 'not(\+ false)') isNil.
	self assert: (Prolog new evaluate: 'not(\+ true)') = #(()).! !

!PrologTest methodsFor: 'definite clause grammar' stamp: 'spfa 10/7/2022 18:21'!
testPalindrome

	self assert: (Metalevel new ? 'phrase(palindrome, [2])') = #(()).
	self assert: (Metalevel new ? 'phrase(palindrome, [2,2])') = #(()).
	self assert: (Metalevel new ? 'phrase(palindrome, [2,1])') isNil.

	self assert: (Metalevel new ? 'phrase(palindrome, [])') = #(()).
	self assert: (Metalevel new ? 'phrase(palindrome, [2,a,2])') = #(()).
	self assert: (Metalevel new ? 'phrase(palindrome, [2,a,a,2])') = #(()).

	self assert: (Metalevel new ? 'phrase(palindrome, [2,a,a,b,2])') isNil! !

!PrologTest methodsFor: 'definite clause grammar' stamp: 'spfa 10/7/2022 18:21'!
testPalindromeExpanded

	self assert: (Metalevel new ? 'phrase(palindromeExpanded, [2])') = #(()).
	self assert: (Metalevel new ? 'phrase(palindromeExpanded, [2,2])') = #(()).
	self assert: (Metalevel new ? 'phrase(palindromeExpanded, [2,1])') isNil.

	self assert: (Metalevel new ? 'phrase(palindromeExpanded, [])') = #(()).
	self assert: (Metalevel new ? 'phrase(palindromeExpanded, [2,a,2])') = #(()).
	self assert: (Metalevel new ? 'phrase(palindromeExpanded, [2,a,a,2])') = #(()).

	self assert: (Metalevel new ? 'phrase(palindromeExpanded, [2,a,a,b,2])') isNil! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/14/2022 14:28'!
testParenthesis

 	self assert: (Prolog new evaluate: '(=(1,1),=(2,2),=(3,3))') = #(()).
	self assert: (Prolog new evaluate: '((=(1,1),=(2,2),=(3,3)))') = #(()).

	self assert: (Prolog new evaluate: '((=(1,1),=(2,2),=(3,4)))') isNil! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/14/2022 15:08'!
testParenthesis2

	self assert: (Prolog new evaluate: 'not((=(1,1),=(1,2)))') = #(()).
	self assert: (Prolog new evaluate: 'not((=(1,2),=(1,1)))') = #(()).! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/14/2022 15:10'!
testParenthesis3

 	self assert: (Prolog new evaluate: 'N is (((4)))') = #(#(4)).

 	self assert: (Prolog new evaluate: 'N is ((2 + (2)))') = #(#(4)).! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/14/2022 15:15'!
testParenthesis4

 	self assert: (Prolog new evaluate: 'N is 4 - 1 - 1 + 2') = #(#(4)).

 	self assert: (Prolog new evaluate: 'N is ((4 - (1 + 1) + 2))') = #(#(4)).

 	self assert: (Prolog new evaluate: 'N is ((4 - (((1)) + 1) + (2)))') = #(#(4)).! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/10/2022 23:19'!
testPrefix
 
	self assert: (Prolog new evaluate: '\+ true') isNil.
	self assert: (Prolog new evaluate: '\+ false') = #(()).! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/10/2022 23:29'!
testPrefix2
 
	self should: [Prolog new evaluate: '\+ X'] raise: Error.
	self assert: (Prolog new evaluate: '=**(\+ X,List).' withBindings: true) 
					printString = '{{List = [\+, X].}}'.! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/12/2022 15:50'!
testPrefix3
 
	self assert: (Prolog new evaluate: '\+ (=(1,1))') isNil.
	self assert: (Prolog new evaluate: '\+ (=(1,2))') = #(()).
! !

!PrologTest methodsFor: 'core' stamp: 'spfa 10/12/2022 16:55'!
testPrefix4
 
	self assert: (Prolog new evaluate: 'not((=(1,1),=(1,2)))') = #(()). 

	self should: [Prolog new evaluate: '\+(=(1,1),=(1,2))'] raise: Error.
	self assert: (Prolog new evaluate: '\+ (=(1,1),=(1,2))') = #(()).
! !

!PrologTest methodsFor: 'interaction' stamp: 'spfa 10/13/2022 16:52'!
testQuery

	| gen |

	gen := Prolog new query: '=(X,[a,b,c]),append(_,[X],Y)'.

	self assert: (gen next: 3) last printString = '{X = [a, b, c]. . Y = [_, _, [a, b, c]].}'.
	self assert: (gen next: 3) last printString = '{X = [a, b, c]. . Y = [_, _, _, _, _, [a, b, c]].}' .! !

!PrologTest methodsFor: 'parsing' stamp: 'spfa 10/20/2022 13:30'!
testReadClause

	| parser terms |

	parser := LParserForTesting reading: 'X = (hello(world), bye(sky)), Y = (bye(world), hello(sky)).'.

	terms := parser readClause. 

	self assert: terms printString = ',(=(''X'', ,(hello(world), bye(sky))), =(''Y'', ,(bye(world), hello(sky))))' ! !

!PrologTest methodsFor: 'parsing' stamp: 'spfa 10/20/2022 12:05'!
testReadClause2

	| parser |

	parser := LParserForTesting reading: '(hello(world), bye(sky)).'.

	self assert: parser readClause printString = ',(hello(world), bye(sky))'.

	parser := LParserForTesting reading: 'hello(world), bye(sky).'.

	self assert: parser readClause printString = ',(hello(world), bye(sky))'! !

!PrologTest methodsFor: 'definite clause grammar' stamp: 'spfa 10/7/2022 18:21'!
testReversal

	self assert: (Metalevel new ? 'phrase(reversal([1,2,3,4,5]),RL)') printString = '{{[5, 4, 3, 2, 1]}}'.

	self assert: (Metalevel new ? 'phrase(reversal([1,2,3]),[3,2,1])') printString = '#(#())'.

	self assert: (Metalevel new ? 'phrase(reversal([1,2,3]),[3,2,2])') isNil! !

!PrologTest methodsFor: 'smalltalk interplay' stamp: 'spfa 10/11/2022 10:50'!
testSend

	| v |

	self assert: (Prolog evaluate: 'send(4,#@,[10],X)') first first = (4@10).
	self assert: (Prolog evaluate: 'send(4,even,X)') = #((true)).
	self assert: (Prolog evaluate: 'send(4,odd,X)') = #((false)).

	self assert: (Prolog evaluate: 'send(4,#even,X)') = #((true)).

	v := Prolog evaluate: 'iss(M, Morph new), iss(P, m extent), iss(B, p logicValue isPoint),test(B)'.
	self assert: v size == 1.
	self assert: v first first isMorph.
	self assert: v first second isPoint.

	v := Prolog evaluate: 'iss(MC,Morph),send(MC,new,M), send(M,extent,P), iss(B, p logicValue isPoint),test(B)'.
	self assert: v size == 1.
	self assert: v first second isMorph.
	self assert: v first third isPoint.
! !

!PrologTest methodsFor: 'higher-order' stamp: 'spfa 10/3/2022 21:45'!
testSetof

	self assert: 
		(PrologILDForTesting evaluate: 'setof(P, likes(P, pizza), S)' getBindings: 'S') 
			last printString = '[mary, marco]'.! !

!PrologTest methodsFor: 'higher-order' stamp: 'spfa 9/28/2022 00:29'!
testSetof2

	self assert: 
		(PrologILDForTesting evaluate: 'setof(person, likes(person, ''wine''), bag)') isNil! !

!PrologTest methodsFor: 'smalltalk interplay' stamp: 'spfa 9/28/2022 15:02'!
testSmalltalkQuery

	"see #asPrologList in Symbol"

	self assert: (Family ? #(father ~ ~)) 
		= #(('John' 'Mary') ('John' 'David') ('David' 'Jack') ('Arthur' 'Nancy')).

	self assert: (Family ? #(father ~ 'David')) 
		= #(('John')).

	self assert: (Family ? #(father 'David' ~)) 
		= #(('Jack')).

	self assert: (Family ? #(father ~ ~~)) 
		= #(('John') ('John') ('David') ('Arthur')).

	self assert: (Family ? #(father ~~ ~~)) 
		= #(() () () ()).

	self assert: (Family ? #(and (grandPa 'John' ~) (mother ~ ~~)) )
		= #(('Nancy' 'Mary') ('Jack' 'Mary'))! !

!PrologTest methodsFor: 'smalltalk interplay' stamp: 'spfa 9/28/2022 15:02'!
testSmalltalkQuery2

	| x y |

	x := LogicRef new.
	y := LogicRef new.
	self assert: (Family ? {#and . {#grandPa . 'John' . x} . {#mother . y . x}} )
		=  #(('Nancy' 'Mary'))! !

!PrologTest methodsFor: 'smalltalk interplay' stamp: 'spfa 9/28/2022 15:02'!
testSmalltalkQuery3

	self assert: (Prolog ? #(member ~ (1 2 3))) = #((1) (2) (3)).
! !

!PrologTest methodsFor: 'complex' stamp: 'spfa 10/3/2022 22:12'!
testTicTacToe

	self assert: (Chapter5 ? 'init,record(''x'',1,1),board(B),findall([X,Y],mark(''o'',B,X,Y),Moves)') 
		first last printString = '[[2, 1], [3, 1], [1, 2], [2, 2], [3, 2], [1, 3], [2, 3], [3, 3]]' ! !

!PrologTest methodsFor: 'complex' stamp: 'spfa 10/4/2022 10:55'!
testTicTacToe2

	self assert: (Chapter5 ? 
			'go,h(2,2),c,h(1,2),c,h(1,3),c,h(3,3),c,board(P),=(P,[o, o, o, x, x, o, x, _, x])') 
		notNil! !

!PrologTest methodsFor: 'complex' stamp: 'spfa 10/4/2022 16:24'!
testTicTacToe3

	self assert: (Chapter5 ? 
			'go,h(2,2),c,h(3,3),c,h(3,1),c,h(1,3),board(B),win(B,x)') 
		notNil! !

!PrologTest methodsFor: 'printing' stamp: 'spfa 10/3/2022 23:49'!
testUnboundPrinting

	self assert: (Prolog ? '=(X,Y),=(X,Z)') printString = '#(#())'.

	self assert: (Prolog evaluate: '=(X,Y),=(X,Z)' withBindings: true) printString = '{{X = Y = Z.}}'.

	self assert: (Prolog evaluate: '=(X,Y),=(X,Z),=(A,B),=(C,B)' withBindings: true) printString
		= '{{X = Y = Z. . A = B = C.}}'.

	self assert: (Prolog evaluate: '=(X,Y),=(X,Z),=(A,B),=(C,B),=(Z,W)' withBindings: true) 
		printString = '{{X = Y = Z = W. . A = B = C.}}'.! !

!PrologTest methodsFor: 'printing' stamp: 'spfa 10/3/2022 23:38'!
testUnboundPrinting2

	self assert: (Prolog ? '=(X,[A,B,C,D]).') printString = '{{[A, B, C, D]}}'.

	self assert: (Prolog evaluate: '=(X,[A,B,C,D]).' withBindings: true) printString
					= '{{X = [A, B, C, D].}}'.! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 22:24'!
testUnification

	self assert: (PrologILDForTesting new ? 
					'retractall(toto),assert(toto([_,C])),toto(B),sayHello(B)')
		printString = '{{''HELLO'' . [_, ''HELLO'']}}'.! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/11/2022 22:49'!
testUniv

	self assert: (Prolog evaluate:'univ(gloup(3, 4),[H|A])') 
					 = {{#gloup . {3 . 4} asPrologList}}.

	self assert: (Prolog evaluate: 'univ(S,[yo,a,b])') 
					= {{{#yo . #a . #b} asPrologRelation}}.

	self assert: (Prolog evaluate: 'univ(yo(U,V),[yo,a,b])') = #((a b)).

	self assert: (Prolog evaluate: 'univ(yo(a,b),[yo,U,V])') = #((a b)).

	self assert: (Prolog evaluate: 'univ(yo(a,b),[yo,a,V])') = #((b)).

	self assert: (Prolog evaluate: 'univ(yo(U,b),[yo,a,V])') = #((a b)).

	self assert: (Prolog evaluate: 'univ(yo(a,U),[yo,V,b])') = #((b a)).

	self assert: (Prolog evaluate: 'univ(yo(U,V),[S,a,b])') = #((a b yo)).
	
	self assert: (Prolog evaluate: 'univ(yo(a,b),[S,a,b])') = #((yo)).

	self assert: (Prolog evaluate: 'univ(yo(U,2),[S,4,V])') = #((4 yo 2)).

	self assert: (Prolog evaluate: 'univ(yo(U,V),[S,a,b])') = #((a b yo)).
		
	self assert: (Prolog evaluate: 'univ(yo(a,b),[S|T])') 
					= {{#yo . {#a . #b} asPrologList}}.

	self assert: (Prolog evaluate: 'univ(yo(a,b),[S,A|T])') 
					= {{#yo . #a . {#b} asPrologList}}.

	self assert: (Prolog evaluate: 'univ(yo(a,b),[yo,a,b])') = #(()).
! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 22:31'!
testUniv2

	"from SWI-Prolog help"

	self assert: (Prolog evaluate:'univ(foo(hello,_),List)') 
					printString = '{{[foo, hello, _]}}'.

	self assert: (Prolog evaluate: 'univ(Term,[baz, foo(1)])')  
					printString = '{{baz(foo(1))}}'! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/3/2022 22:32'!
testUniv3

	self assert: (Prolog evaluate:'=(B,[X,Y,Z]),univ(Term, [bof,B]),=([X,Y,Z],[1,2,3])') 
					printString = '{{[1, 2, 3] . 1 . 2 . 3 . bof([1, 2, 3])}}'! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/10/2022 15:03'!
testUniv4

	self assert: (Prolog evaluate:'univ(=(X,1),L),univ(F,L),call(F)') 
					printString = '{{1 . [=, 1, 1] . =(1, 1)}}'.! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/11/2022 11:07'!
testUniv5

	self assert: (Prolog evaluate:'univ(is(X,1+1),L),univ(F,L),call(F)') 
					printString = '{{2 . [is, 2, +(1, 1)] . is(2, +(1, 1))}}'.

	self assert: (Prolog evaluate:'univ(iss(X,1+1),L),univ(F,L),call(F)') 
					printString = '{{2 . [iss, 2, 1 + 1] . iss(2, 1 + 1)}}'.! !

!PrologTest methodsFor: 'predicates' stamp: 'spfa 10/8/2022 00:31'!
testVar

	self assert: (Prolog evaluate: 'var(G)') notNil.
	self assert: (Prolog evaluate: 'var(_)') notNil.

	self assert: (Prolog evaluate: 'nonvar(G)') isNil.
	self assert: (Prolog evaluate: 'nonvar(_)') isNil.

	self assert: (Prolog evaluate: 'var(X),=(X,f(_)),nonvar(X)') notNil.
	self assert: (Prolog evaluate: 'var(X),=(X,_),nonvar(X)') isNil.

	self assert: (Prolog evaluate: '=(X,Y),var(Y)') notNil.! !
